"""
Contains TmProc class for calculating TURBOMOLE related properties of conformers.
"""

import os
import shutil
import math
from pathlib import Path
from typing import cast, final, override

from .qm_processor import QmProc
from ..logging import setup_logger
from .job import (
    JobContext,
)
from .results import (
    GsolvResult,
    MetaData,
    NMRResult,
    RotResult,
    OptResult,
    SPResult,
)
from ..params import (
    WARNLEN,
    R,
    AU2KCAL,
    TmSolvMod,
    ASSETS_PATH,
    Prog,
    ENVIRON,
    USER_ASSETS_PATH,
)
from ..utilities import Factory
from ..config.job_config import NMRJobConfig, SPJobConfig, XTBOptJobConfig, RotJobConfig
from ..assets import FUNCTIONALS, SOLVENTS

logger = setup_logger(__name__)


@final
class TmProc(QmProc):
    """
    Performs calculations using TURBOMOLE.
    """

    progname = Prog.TM

    __gridsettings = {
        "low": ["    gridsize m3", "$scfconv 6"],
        "low+": ["    gridsize m4", "$scfconv 6"],
        "high": ["    gridsize m4", "$scfconv 7"],
        "high+": ["    gridsize m5", "$scfconv 7"],
        "nmr": ["    gridsize 5", "$scfconv 7"],
    }

    __returncode_to_err: dict[int, str] = {}

    # COSMO dielectric constants stored here
    # (mapped to solvent names under 'cosmo' in the solvent helper)
    __cosmo_dcs = {
        "acetone": 20.7,
        "acetonitrile": 36.6,
        "aniline": 6.9,
        "benzaldehyde": 18.2,
        "benzene": 2.3,
        "ccl4": 2.2,
        "ch2cl2": 9.1,
        "chcl3": 4.8,
        "cs2": 2.6,
        "cyclohexane": 2.0,
        "dichloroethane": 10.125,
        "diethylether": 4.4,
        "dioxane": 2.2,
        "dmf": 38.3,
        "dmso": 47.2,
        "ethanol": 24.6,
        "ethylacetate": 5.9,
        "furan": 3.0,
        "h2o": 80.1,
        "hexadecane": 2.1,
        "hexane": 1.9,
        "methanol": 32.7,
        "nitromethane": 38.2,
        "octane": 1.94,
        "octanol": 9.9,
        "phenol": 8.0,
        "thf": 7.6,
        "toluene": 2.4,
        "woctanol": 8.1,
    }

    # Contains mapping from lower case to proper spelling of bases for TM
    # (should be available for most elements)
    # TODO: consider adding custom basis mapping in user assets
    __basis_mapping = {
        "dzp": "DZP",
        "svp": "SVP",
        "def-svp": "def-SVP",
        "def2-svp": "def2-SVP",
        "dhf-svp": "dhf-SVP",
        "dhf-svp-2c": "dhf-SVP-2c",
        "tz": "TZ",
        "tzv": "TZV",
        "tzp": "TZP",
        "tzvp": "TZVP",
        "def-tzvp": "def-TZVP",
        "def2-tzvp": "def2-TZVP",
        "dhf-tzvp": "dhf-TZVP",
        "dhf-tzvp-2c": "dhf-TZVP-2c",
        "tzvpp": "TZVPP",
        "def-tzvpp": "def-TZVPP",
        "def2-tzvpp": "def2-TZVPP",
        "def2-mtzvpp": "def2-mTZVPP",
        "dhf-tzvpp": "dhf-TZVPP",
        "dhf-tzvpp-2c": "dhf-TZVPP-2c",
        "tzvppp": "TZVPPP",
        "def-qzv": "def-QZV",
        "def2-qzv": "def2-QZV",
        "qzv": "QZV",
        "qz": "QZ",
        "def-qzvp": "def-QZVP",
        "def2-qzvp": "def2-QZVP",
        "dhf-qzvp": "dhf-QZVP",
        "dhf-qzvp-2c": "dhf-QZVP-2c",
        "qzvp": "QZVP",
        "qzp": "QZP",
        "def-qzvpp": "def-QZVPP",
        "def2-qzvpp": "def2-QZVPP",
        "dhf-qzvpp": "dhf-QZVPP",
        "dhf-qzvpp-2c": "dhf-QZVPP-2c",
        "qzvpp": "QZVPP",
        "qzpp": "QZPP",
        "minix": "minix",
        "sto-3g": "sto-3g",
        "3-21g": "3-21g",
        "4-31g": "4-31g",
        "def2-qzvppd": "def2-QZVPPD",
        "def2-qzvpd": "def2-QZVPD",
        "6-31g": "6-31G",
        "6-31g*": "6-31G*",
        "6-31g**": "6-31G**",
        "6-311g": "6-311G",
        "6-311g*": "6-311G*",
        "6-311g**": "6-311G**",
        "6-311++g**": "6-311++G**",
        "6-311g(2df,2pd)": "6-311G(2df,2pd)",
        "cc-pvdz": "cc-pVDZ",
        "aug-cc-pvdz": "aug-cc-pVDZ",
        "yp-aug-cc-pvdz": "YP-aug-cc-pVDZ",
        "d-aug-cc-pvdz": "d-aug-cc-pVDZ",
        "cc-pvtz": "cc-pVTZ",
        "cc-pvtz-kernel": "cc-pVTZ-kernel",
        "aug-cc-pvtz": "aug-cc-pVTZ",
        "yp-aug-cc-pvtz": "YP-aug-cc-pVTZ",
        "d-aug-cc-pvtz": "d-aug-cc-pVTZ",
        "d-aug-cc-pvtz-oep": "d-aug-cc-pVTZ-oep",
        "cc-pvqz": "cc-pVQZ",
        "aug-cc-pvqz": "aug-cc-pVQZ",
        "yp-aug-cc-pvqz": "YP-aug-cc-pVQZ",
        "d-aug-cc-pvqz": "d-aug-cc-pVQZ",
        "cc-pv5z": "cc-pV5Z",
        "aug-cc-pv5z": "aug-cc-pV5Z",
        "yp-aug-cc-pv5z": "YP-aug-cc-pV5Z",
        "d-aug-cc-pv5z": "d-aug-cc-pV5Z",
        "cc-pv6z": "cc-pV6Z",
        "aug-cc-pv6z": "aug-cc-pV6Z",
        "r12": "r12",
        "cc-pvdz-f12": "cc-pVDZ-F12",
        "cc-pvtz-f12": "cc-pVTZ-F12",
        "cc-pvqz-f12": "cc-pVQZ-F12",
        "def2-svpd": "def2-SVPD",
        "def2-tzvppd": "def2-TZVPPD",
        "def2-tzvpd": "def2-TZVPD",
        "dz": "DZ",
        "sv": "SV",
        "sv(p)": "SV(P)",
        "def-sv(p)": "def-SV(P)",
        "def2-sv(p)": "def2-SV(P)",
        "dhf-sv(p)": "dhf-SV(P)",
        "dhf-sv(p)-2c": "dhf-SV(P)-2c",
    }

    def __prep(
        self,
        job: JobContext,
        config: SPJobConfig,
        jobtype: str,
        jobdir: str | Path,
        no_solv: bool = False,
    ) -> None:
        """
        Prepares TURBOMOLE input files for a specified jobtype.

        :param job: Job context
        :type job: JobContext
        :param config: Configuration for the job
        :type config: SPJobConfig
        :param jobtype: Specified jobtype
        :type jobtype: str
        :param jobdir: Path to the job directory
        :type jobdir: str | Path
        :param no_solv: If True, no solvent model is used
        :type no_solv: bool
        :returns: None
        :rtype: None
        """
        inp = []

        func = config.func
        func_name = FUNCTIONALS[func][Prog.TM.value]
        func_type = FUNCTIONALS[func]["type"]

        try:
            basis = self.__basis_mapping[config.basis.lower()]
        except KeyError as exc:
            raise KeyError(
                f"Basis {config.basis} could not be found for TURBOMOLE input preparation. "
                f"Available basis sets: {list(self.__basis_mapping.values())}"
            ) from exc

        inp.extend(["$atoms", f"    basis={basis}"])

        if basis.endswith("D"):
            inp.append("    jbas=universal")

        inp.extend(["$dft", f"   functional {func_name}"])

        # Configure grid
        inp.extend(self.__gridsettings[config.grid])

        # r2scan-3c should use m4 grid and radsize 10
        if func == "r2scan-3c":
            if "m3" in inp:
                inp[inp.index("m3")] = "m4"
            inp.insert(inp.index("$dft") + 2, "    radsize 10")

        # Add dispersion
        # dispersion correction information
        # TODO: temporary solution (not very nice)
        mapping = {
            "d3bj": "$disp3 -bj",
            "d3(0)": "$disp3 -zero",
            "d4": "$disp4",
            "nl": "$donl",
        }

        disp = FUNCTIONALS[func]["disp"]
        if disp not in ["composite", "novdw", "included"]:
            inp.append(mapping[disp])
        elif func.endswith("-v"):
            inp.append("$doscnl")

        inp.append("$rij")

        # Add charge and unpaired info
        inp.append(f"$charge={job.charge} unpaired={job.unpaired}")

        # Write coord file
        with open(os.path.join(jobdir, "coord"), "w") as f:
            f.writelines(job.conf.tocoord())

        inp.append("$coord file=coord")

        # Do further manipulations of input files
        self.__prep_main(
            inp,
            func,
            disp,
            func_type,
            basis,
        )
        if not no_solv:
            self.__prep_solv(inp, config, jobtype)

        self.__prep_special(inp, config, jobtype)

        # Append template
        if config.template:
            template_path = Path(USER_ASSETS_PATH / f"{job.from_part}.tm.template")
            if template_path.is_file():
                lines = template_path.read_text().split("\n")
                inp.extend(lines)

        if jobtype == "xtb_opt":
            inp.append("$grad file=gradient")

        inp.append("$end")
        inp = [i if i.endswith("\n") else i + "\n" for i in inp]
        (Path(jobdir) / "control").write_text("".join(inp))

    def __prep_main(
        self,
        inp: list[str],
        func: str,
        disp: str,
        func_type: str,
        basis: str,
    ):
        # Special treatment for KT1/KT2
        if "kt" in func:
            func_line_index = next(
                inp.index(line) for line in inp if "functional" in line
            )
            inp[func_line_index] = "   functional xcfun set-gga\n"
            inp.insert(func_line_index + 1, f"   functional xcfun kt{func[2]} 1.0\n")
        # Special treatment for b97-3c
        elif func == "b97-3c":
            # Needs three-body dispersion
            disp_line_index = next(inp.index(line) for line in inp if "disp" in line)
            inp[disp_line_index] = "$disp3 -bj -abc\n"

        # Enable non local dispersion
        if disp == "nl":
            inp.append("$donl\n")

        # Handle GCP
        if func_type != "composite":
            gcp_keywords = {
                "minis": "MINIS",
                "sv": "SV",
                "6-31g(d)": "631GD",
                "def2-sv(p)": "SV(P)",
                "def2-svp": "SVP",
                "def2-tzvp": "TZ",
            }
            if basis.lower() in gcp_keywords:
                if basis.lower() == "def2-sv(p)":
                    inp.append("$gcp dft/sv(p)\n")
                else:
                    inp.append(f"$gcp dft/{basis.lower().replace('-', '')}\n")

    def __prep_solv(self, inp: list[str], config: SPJobConfig, jobtype: str):
        inp.insert(-1, "$cosmo\n")

        assert config.solvent
        assert config.sm

        sm = config.sm
        solv_key = SOLVENTS[config.solvent][sm]

        # write DC in any case
        inp.append(f" epsilon= {self.__cosmo_dcs[solv_key]}\n")

        if sm == TmSolvMod.DCOSMORS:
            # if using dcosmors also add the potential file path
            if solv_key not in [
                "woctanol",
                "hexadecane",
                "octanol",
            ]:
                inp.append(
                    f"$dcosmo_rs file={solv_key}_25.pot\n",
                )
            else:
                # The three solvents above are specifically defined in the assets
                # TODO: this opens the possibility to insert your own potential files
                inp.append(
                    f"$dcosmo_rs file={os.path.join(ASSETS_PATH, solv_key)}_25.pot\n",
                )

        if jobtype == "rot":
            inp.extend(
                [
                    " cavity closed\n",
                    " use_contcav\n",
                    " nspa=272\n",
                    " nsph=162\n",
                    "$cosmo_isorad\n",
                ]
            )

    def __prep_special(self, lines: list[str], config: SPJobConfig, jobtype: str):
        # Set NMR parameters
        if "nmr" in jobtype:
            assert isinstance(config, NMRJobConfig)
            # Determine the settings that need to be put into the input file for the NMR calculation
            # $nucsel does not work properly with capital letters
            todo = [element.lower() for element in config.active_nuclei.split(",")]

            lines.append("$rpacor 10000")

            lines.append("$ncoupling")

            if config.fc_only:
                lines.extend([" simple\n", " thr=0.0\n"])

            # nucsel only required if not all elements are active
            if not all(element in todo for element in ["h", "c", "f", "si", "p"]):
                todo = [f'"{e}"' for e in todo]
                lines.extend(
                    [
                        "$nucsel " + " ".join(todo) + "\n",
                        "$nucsel2 " + " ".join(todo) + "\n",
                    ]
                )

            lines.append("$rpaconv 8\n")
        elif jobtype == "rot":
            assert isinstance(config, RotJobConfig)
            lines.extend(
                ["$scfinstab dynpol nm"]
                + [f" {i}" for i in config.freq]
                + ["$velocity gauge", "$rpaconv 4"]
            )

    @staticmethod
    def __copy_mo(
        jobdir: str, guess_file: str | Path | tuple[str | Path, str | Path]
    ) -> None:
        """
        Copy the MO file(s) for TURBOMOLE (should be TM format).

        :param jobdir: Path to the job directory
        :type jobdir: str
        :param guess_file: Path to the guess file(s)
        :type guess_file: str | Path | tuple[str | Path, str | Path]
        :returns: None
        :rtype: None
        """
        if isinstance(guess_file, tuple):
            # open shell guess
            if all(
                os.path.isfile(f)
                and any(g in str(f) for g in ["alpha", "beta"])
                and not any(
                    os.path.join(jobdir, g) == str(f) for g in ["alpha", "beta"]
                )
                for f in guess_file
            ):
                # All MO files found and not already in dir
                # Copy MO files
                for f in guess_file:
                    g = os.path.split(f)[1] if isinstance(f, str) else f.stem
                    logger.debug(
                        f"{f'worker{os.getpid()}:':{WARNLEN}}Copying {g} file from {f}."
                    )
                    shutil.copy(f, os.path.join(jobdir, g))
        else:
            # closed shell guess
            if (
                os.path.isfile(guess_file)
                and os.path.split(guess_file)[1] == "mos"
                and os.path.join(jobdir, "mos") != guess_file
            ):
                # Copy MO file
                logger.debug(
                    f"{f'worker{os.getpid()}:':{WARNLEN}}Copying mos file from {guess_file}."
                )
                shutil.copy(guess_file, os.path.join(jobdir, "mos"))

    @staticmethod
    def __check_output(lines: list[str]) -> str:
        """
        Checks the lines from the output file for errors and returns them.

        :param lines: list of lines from the output file.
        :type lines: list[str]
        :returns: error message if an error was found, empty string otherwise
        :rtype: str
        """
        # Dict mapping specific messages from the output to error messages
        # TODO: this should be extended later
        out_to_err = {
            "SCF NOT CONVERGED": "scf_not_converged",
        }
        for line in lines:
            if any(key in line for key in out_to_err.keys()):
                # Returns the first error found
                key = next(filter(lambda x: x in line, out_to_err.keys()))
                return out_to_err[key]
        return ""

    def _sp(
        self,
        job: JobContext,
        config: SPJobConfig,
        jobdir: str | Path | None = None,
        no_solv: bool = False,
        prep: bool = True,
    ) -> tuple[SPResult, MetaData]:
        """
        TURBOMOLE single-point calculation.

        :param job: JobContext object containing the job information, metadata is stored in job.meta
        :type job: JobContext
        :param config: Configuration for the job
        :type config: SPJobConfig
        :param jobdir: path to the job directory
        :type jobdir: str | Path | None
        :param no_solv: if True, no solvent model is used
        :type no_solv: bool
        :param prep: if True, a new input file is generated
        :type prep: bool
        :returns: Tuple of (SPResult, MetaData)
        :rtype: tuple[SPResult, MetaData]
        """
        if jobdir is None:
            jobdir = self._setup(job, "sp")

        # set results
        result = SPResult()

        meta = MetaData(job.conf.name)

        # set in/out path
        outputpath = os.path.join(jobdir, "ridft.out")

        # check, if there is an existing mo/alpha,beta file and copy it if option
        # 'copy_mo' is true
        # mo files: mos/alpha,beta
        # NOTE: this HAS TO BE in this order, otherwise ridft fails to read mos
        if config.copy_mo and job.mo_guess is not None:
            self.__copy_mo(str(jobdir), job.mo_guess)

        if prep:
            self.__prep(job, config, "sp", jobdir, no_solv=config.gas_phase or no_solv)

        # call turbomole
        call = [str(Path(config.paths.tm) / "ridft")]
        env = ENVIRON.copy()
        env["PARA_ARCH"] = "SMP"
        env["PARNODES"] = str(job.omp)
        returncode, _ = self._make_call(call, outputpath, jobdir, env=env)

        meta.success = returncode == 0

        lines = Path(outputpath).read_text().split("\n")

        # Check for errors in the output file in case returncode is 0
        if meta.success:
            meta.error = self.__check_output(lines)
            meta.success = meta.error == ""
        else:
            meta.error = self.__returncode_to_err.get(returncode, "unknown_error")

        # Get final energy
        try:
            result.energy = next(
                float(line.split()[4])
                for line in lines
                if "|  total energy      = " in line
            )
        except StopIteration:
            meta.success = False
            meta.error = "Could not parse final energy"

        if config.copy_mo:
            # store the path to the current MO file(s) for this conformer if possible
            if os.path.isfile(os.path.join(jobdir, "mos")):
                result.mo_path = os.path.join(jobdir, "mos")
            elif os.path.isfile(os.path.join(jobdir, "alpha")):
                result.mo_path = (
                    os.path.join(jobdir, "alpha"),
                    os.path.join(jobdir, "beta"),
                )

        return result, meta

    @override
    def sp(self, *args, **kwargs):
        """
        Perform single-point calculation.

        :param args: Arguments.
        :param kwargs: Keyword arguments.
        :return: Tuple of (SP result, metadata).
        """
        return self._sp(*args, **kwargs)

    @override
    def gsolv(
        self,
        job: JobContext,
        config: SPJobConfig,
    ) -> tuple[GsolvResult, MetaData]:
        """
        Calculate the solvation contribution to the free enthalpy explicitely using (D)COSMO(RS).

        :param job: JobContext object containing the job information, metadata is stored in job.meta
        :param config: SP configuration
        :return: Tuple of (GsolvResult, MetaData)
        """
        # what is returned in the end
        result = GsolvResult()
        meta = MetaData(job.conf.name)

        jobdir = self._setup(job, "gsolv")

        if not config.sm or not config.solvent or not config.temperature:
            raise ValueError("sm, solvent, or temperature not set.")
        # assert config.multitemp
        # assert config.trange

        if config.sm in [TmSolvMod.COSMO, TmSolvMod.DCOSMORS]:
            # Non-COSMORS procedure:
            # Run gas-phase sp
            spres, spmeta = self._sp(job, config, jobdir=jobdir, no_solv=True)

            if spmeta.success:
                result.energy_gas = spres.energy
            else:
                meta.success = False
                meta.error = spmeta.error
                return result, meta

            # Run solution sp
            spres, spmeta = self._sp(job, config, jobdir=jobdir)

            if spmeta.success:
                result.energy_solv = spres.energy
            else:
                meta.success = False
                meta.error = spmeta.error
                return result, meta

            # Calculate gsolv from energy difference
            result.gsolv = result.energy_solv - result.energy_gas
            meta.success = True
        else:
            # COSMORS procedure:
            # Run gas-phase sp with unaltered settings
            spres, spmeta = self._sp(job, config, jobdir=jobdir, no_solv=True)

            if spmeta.success:
                result.energy_gas = spres.energy
            else:
                meta.success = False
                meta.error = spmeta.error
                return result, meta

            # Run gas-phase sp with BP86 and def2-TZVP (normal)/def2-TZVPD (fine)
            # Turn off copying mos since this will lead to errors otherwise
            # (due to incorrect order of prep/copy_mos)
            spconfig = SPJobConfig(
                copy_mo=False,
                func="bp86",
                basis="def2-tzvp" if config.sm == TmSolvMod.COSMORS else "def2-tzvpd",
                grid=config.grid,
                template=config.template,
                gas_phase=True,
                paths=config.paths,
            )

            spres, spmeta = self._sp(job, spconfig, jobdir=jobdir, no_solv=True)

            if not spmeta.success:
                meta.success = False
                meta.error = spmeta.error
                return result, meta

            # Run special cosmo sp with BP86 and def2-TZVP (normal)/def2-TZVPD (fine)
            # Write special settings for cosmo into control file
            with open(os.path.join(jobdir, "control"), "r+") as f:
                lines = f.readlines()

                lines[-1:-1] = [
                    "$cosmo\n",
                    " epsilon=infinity\n",
                    " use_contcav\n",
                    " cavity closed\n",
                    " nspa=272\n",
                    " nsph=162\n",
                    "$cosmo_out  file=out.cosmo\n",
                ]

                f.seek(0)
                f.writelines(lines)
                f.truncate()

            # Run sp
            spconfig.gas_phase = True
            spconfig.sm = config.sm
            spconfig.solvent = config.solvent
            spres, spmeta = self._sp(job, spconfig, jobdir=jobdir, prep=False)

            if not spmeta.success:
                meta.success = False
                meta.error = spmeta.error
                return result, meta

            # Prepare cosmotherm.inp
            if config.sm == TmSolvMod.COSMORS:
                setup = (
                    config.paths.cosmorssetup
                    if "FINE" not in config.paths.cosmorssetup
                    else config.paths.cosmorssetup.replace("TZVPD_FINE", "TZVP")
                )
            else:
                setup = (
                    config.paths.cosmorssetup
                    if "FINE" in config.paths.cosmorssetup
                    else config.paths.cosmorssetup.replace("TZVP", "TZVPD_FINE")
                )
            lines = [
                f"ctd = {setup} cdir = {(Path(config.paths.cosmotherm).parent / '..' / 'CTDATA-FILES').resolve()}\n",
                "EFILE VPFILE\n",
                "!!\n",
            ]
            db = (
                Path(config.paths.cosmotherm).parent
                / ".."
                / "DATABASE-COSMO"
                / (
                    "BP-TZVP-COSMO"
                    if config.sm == TmSolvMod.COSMORS
                    else "BP-TZVPD-FINE"
                )
            ).resolve()

            solv_key = SOLVENTS[config.solvent][config.sm]
            if solv_key == "woctanol":
                lines.extend(
                    [
                        f"f = h2o.cosmo fdir={db} autoc\n",
                        f"f = 1-octanol.cosmo fdir={db} autoc\n",
                    ]
                )
                mix = "0.27 0.73"
            else:
                lines.append(f"f = {solv_key}.cosmo" + f" fdir={db} autoc\n")
                mix = "1.0 0.0"

            lines.append("f = out.cosmo\n")

            # TODO: are these temperature dependent terms used somewhere?
            # if config.multitemp:
            #     trange = frange(
            #         config.trange[0],
            #         config.trange[1],
            #         step=config.trange[2],
            #     )
            #
            #     # Always append the fixed temperature to the trange so that it is the last value
            #     trange.append(config.temperature)
            #
            #     # Write trange to the xcontrol file
            #     for t in trange:
            #         lines.append(f"henry xh={{{mix}}} tc={t - 273.15} Gsolv\n")
            # else:
            lines.append(f"henry xh={{{mix}}} tc={config.temperature - 273.15} Gsolv\n")

            with open(os.path.join(jobdir, "cosmotherm.inp"), "w") as f:
                f.writelines(lines)

            # Run cosmotherm
            outputpath = os.path.join(jobdir, "cosmotherm.out")
            call = [config.paths.cosmotherm, "cosmotherm.inp"]
            env = ENVIRON.copy()
            env["PARA_ARCH"] = "SMP"
            env["PARNODES"] = str(job.omp)
            returncode, _ = self._make_call(call, outputpath, jobdir, env=env)

            meta.success = returncode == 0
            if not meta.success:
                meta.error = "unknown_error"
                return result, meta

            # Read output
            gsolvt = {}
            videal = (
                24.789561955 / 298.15
            )  # molar volume for ideal gas at 298.15 K 100.0 kPa

            cosmothermtab = os.path.join(jobdir, "cosmotherm.tab")
            with open(cosmothermtab) as inp:
                lines = inp.readlines()
            temp = 0.0
            vwork = 0.0
            for line in lines:
                if "T=" in line:
                    temp = float(line.split()[5])

                    # Add volume work
                    vwork = R * temp * math.log(videal * temp)
                elif " out " in line:
                    # Add volume work
                    gsolvt[temp] = float(line.split()[-1]) / AU2KCAL + vwork / AU2KCAL

            # result.gsolvt = gsolvt
            if config.temperature not in gsolvt:
                logger.warning(
                    f"Job for {job.conf.name} failed. Gsolv not parsed correcly."
                )
                meta.error = "parsing_error"
                return result, meta
            result.gsolv = gsolvt[config.temperature]
            result.energy_solv = result.energy_gas + result.gsolv

            # cosmothermd
            with open(os.path.join(jobdir, "cosmors.out"), "w") as out:
                temp = config.temperature
                vwork = R * temp * math.log(videal * temp)

                out.writelines(
                    [
                        "This is cosmothermrd (python version in ENSO) (SG,FB,SAW, 06/18)\n",
                        "final thermochemical solvation properties in kcal/mol\n"
                        "----------------------------------------------------------\n",
                        " Gsolv({} K)= {:10.3f}\n".format(
                            temp, result.gsolv * AU2KCAL - vwork
                        ),
                        " VWork({} K)= {:10.3f}\n".format(temp, vwork),
                        " Gsolv+VWork({} K)= {:10.3f}\n".format(
                            # volwork already included!
                            temp,
                            result.gsolv * AU2KCAL,
                        ),
                    ]
                )

        return result, meta

    @override
    def xtb_opt(
        self,
        job: JobContext,
        config: XTBOptJobConfig,
        filename: str = "xtb_opt",
    ) -> tuple[OptResult, MetaData]:
        """
        Geometry optimization using ANCOPT and ORCA gradients.
        Note that solvation is handled here always implicitly.

        :param job: JobContext object containing the job information, metadata is stored in job.meta
        :param config: XTB optimization configuration
        :return: Tuple of (OptResult, MetaData)
        """
        result = OptResult()
        meta = MetaData(job.conf.name)

        jobdir = self._setup(job, "xtb_opt")
        filename = "xtb_opt"

        xcontrolname = "xtb_opt-xcontrol-inp"

        files = [
            "xtbrestart",
            "xtbtopo.mol",
            xcontrolname,
            "wbo",
            "charges",
            "gfnff_topo",
        ]

        # remove potentially preexisting files to avoid confusion
        for file in files:
            if os.path.isfile(os.path.join(jobdir, file)):
                os.remove(os.path.join(jobdir, file))

        # prepare configuration file for ancopt (xcontrol file)
        with open(os.path.join(jobdir, xcontrolname), "w", newline=None) as out:
            out.write("$opt \n")
            if config.macrocycles:
                out.write(f"maxcycle={config.optcycles} \n")
                out.write(f"microcycle={config.optcycles} \n")

            out.writelines(
                [
                    "average conv=true \n",
                    f"hlow={config.hlow} \n",
                    "s6=30.00 \n",
                    "engine=lbfgs\n",
                ]
            )

            # Import constraints
            if config.constraints:
                out.writelines(config.constraints)

            out.write("$end \n")

        # check, if there is an existing mo/alpha,beta file and copy it if option
        # 'copy_mo' is true
        # mo files: mos/alpha,beta
        if config.copy_mo and job.mo_guess is not None:
            self.__copy_mo(str(jobdir), job.mo_guess)

        self.__prep(job, config, "xtb_opt", jobdir)

        # prepare xtb call
        call = [
            config.paths.xtb,
            "coord",  # name of the coord file generated above
            "--opt",
            config.optlevel,
            "--tm",
            "-I",
            xcontrolname,
        ]

        # set path to the ancopt output file
        outputpath = os.path.join(jobdir, f"{filename}.out")

        # call xtb
        env = ENVIRON.copy()
        env["PARA_ARCH"] = "SMP"
        env["PARNODES"] = str(job.omp)
        returncode, _ = self._make_call(call, outputpath, jobdir, env=env)

        # check if optimization finished without errors
        # NOTE: right now, not converging scfs are not handled because returncodes need to be implemented first
        if returncode != 0:
            meta.success = False
            meta.error = "unknown_error"
            # TODO: the xtb returncodes should be handled
            return result, meta

        # read output
        with open(outputpath) as f:
            lines = f.readlines()

        result.ecyc = []
        result.cycles = 0

        # Substrings indicating error in xtb
        error_ind = [
            "external code error",
            "|grad| > 500, something is totally wrong!",
            "abnormal termination of xtb",
        ]

        # Check if xtb terminated normally (if there are any error indicators
        # in the output)
        meta.success = (
            False
            if next((x for x in lines if any(y in x for y in error_ind)), None)
            is not None
            else True
        )
        if not meta.success:
            meta.error = "unknown_error"
            return result, meta

        # check convergence
        if (
            next((True for x in lines if "GEOMETRY OPTIMIZATION CONVERGED" in x), None)
            is True
        ):
            result.converged = True
        elif (
            next((True for x in lines if "FAILED TO CONVERGE GEOMETRY" in x), None)
            is True
        ):
            result.converged = False

        # Get the number of cycles
        # tmp is one of the values from the dict defined below
        tmp_val = {
            True: ("GEOMETRY OPTIMIZATION CONVERGED", 5),
            False: ("FAILED TO CONVERGE GEOMETRY", 7),
        }[result.converged]

        result.cycles = int(
            next(x for x in lines if tmp_val[0] in x).split()[tmp_val[1]]
        )

        # Get energies for each cycle
        result.ecyc.extend(
            float(line.split("->")[-1])
            for line in filter(lambda x: "av. E: " in x, lines)
        )

        # Get all gradient norms for evaluation
        result.gncyc = [
            float(line.split()[3])
            for line in filter(lambda x: " gradient norm " in x, lines)
        ]

        # Get the last gradient norm
        result.grad_norm = result.gncyc[-1]

        # store the final energy of the optimization in 'energy'
        result.energy = result.ecyc[-1]
        meta.success = True

        if config.copy_mo:
            # store the path to the current MO file(s) for this conformer if possible
            if os.path.isfile(os.path.join(jobdir, "mos")):
                result.mo_path = os.path.join(jobdir, "mos")
            elif os.path.isfile(os.path.join(jobdir, "alpha")):
                result.mo_path = (
                    os.path.join(jobdir, "alpha"),
                    os.path.join(jobdir, "beta"),
                )

        # read out optimized geometry and update conformer geometry with this
        job.conf.fromcoord(os.path.join(jobdir, "xtbopt.coord"))
        result.geom = job.conf.xyz

        return result, meta

    @override
    def opt(self, *args, **kwargs):
        """
        Perform geometry optimization.

        :param args: Arguments.
        :param kwargs: Keyword arguments.
        :return: Tuple of (OptResult, MetaData).
        """
        raise NotImplementedError(
            "Pure TURBOMOLE geometry optimization not available yet."
        )

    @override
    def nmr(
        self,
        job: JobContext,
        config: NMRJobConfig,
    ) -> tuple[NMRResult, MetaData]:
        """
        Calculate the NMR shieldings and/or couplings for a conformer using TURBOMOLE.
        Formatting:
            'shielding' contains a list of tuples (atom_index, shielding), with atom_index being the index of the atom
            in the internal coordinates of the GeometryData.
            'couplings' contains a list of tuples ((atom_index1, atom_index2), coupling), with the indices of the atoms
            in the internal coordinates of the GeometryData. A set is used to represent an atom pair and then converted
            to tuple to be serializable.

        :param job: JobContext object containing the job information, metadata is stored in job.meta
        :param config: NMR configuration
        :return: Tuple of (NMRResult, MetaData)
        """
        # Set results
        result = NMRResult()
        meta = MetaData(job.conf.name)

        jobdir = self._setup(job, "nmr")

        # Run sp first
        self.__prep(job, config, "nmr", jobdir)
        _, spmeta = self._sp(job, config, jobdir=jobdir, prep=False)

        if not spmeta.success:
            meta.success = False
            meta.error = spmeta.error
            return result, meta

        # result.energy = spres.energy

        # Run shielding calculations if requested
        if config.shieldings:
            # Set in/out path
            outputpath = os.path.join(jobdir, "mpshift.out")

            call = [
                str(Path(config.paths.tm) / "mpshift"),
                "-smpcpus",
                f"{job.omp}",
            ]

            # Run mpshift for shielding calculation
            env = ENVIRON.copy()
            env["PARA_ARCH"] = "SMP"
            env["PARNODES"] = str(job.omp)
            returncode, _ = self._make_call(call, outputpath, jobdir, env=env)

            meta.success = returncode == 0
            if not meta.success:
                meta.error = "unknown_error"
                return result, meta

            # Grab shieldings from the output
            with open(outputpath) as f:
                lines = f.readlines()

            try:
                start = lines.index(
                    next(x for x in lines if ">>>>> DFT MAGNETIC SHIELDINGS <<<<<" in x)
                )
            except StopIteration:
                meta.success = False
                meta.error = "Could not read shieldings"
                return result, meta

            lines = lines[start:]

            result.shieldings = []

            # Get lines with "ATOM" in it
            line_indices = [lines.index(line) for line in lines if "ATOM" in line]

            for i in line_indices:
                split = lines[i].split()
                result.shieldings.append((int(split[2]), float(split[4])))

            # Sort shieldings by atom index
            result.shieldings.sort(key=lambda x: x[0])

        # Run couplings calculation if requested
        if config.couplings:
            # Set in/out path
            outputpath = os.path.join(jobdir, "escf.out")

            call = [
                str(Path(config.paths.tm) / "escf"),
                "-smpcpus",
                f"{job.omp}",
            ]

            # Run escf for couplings calculation
            env = ENVIRON.copy()
            env["PARA_ARCH"] = "SMP"
            env["PARNODES"] = str(job.omp)
            returncode, _ = self._make_call(call, outputpath, jobdir, env=env)

            meta.success = returncode == 0
            if not meta.success:
                meta.error = "unknown_error"
                return result, meta

            # Grab couplings from the output
            with open(outputpath) as f:
                lines = f.readlines()

            try:
                start = (
                    lines.index(
                        next(x for x in lines if "Nuclear coupling constants" in x)
                    )
                    + 3
                )
            except StopIteration:
                meta.success = False
                meta.error = "Could not read couplings"
                return result, meta

            lines = lines[start:]

            end = lines.index(next(x for x in lines if "-----" in x))

            lines = lines[:end]

            line_indices = [
                lines.index(line) for line in lines if len(line.split()) in [6, 7]
            ]

            couplings: list[tuple[frozenset[int], float]] = []
            for i in line_indices:
                # pair needs to be a frozenset because normal sets are not hashable and can therefore not be part
                # of a normal set
                split = lines[i].split()
                pair = frozenset((int(split[1]), int(split[4].split(":")[0])))
                coupling = float(split[5])
                couplings.append((pair, coupling))

            # Convert to set and back to get rid of duplicates
            # ('vectorizing the symmetric matrix')
            couplings = list(set(couplings))

            # Convert all the frozensets to a tuple to be serializable
            for i in range(len(couplings)):
                pair_tuple = cast(tuple[int, int], tuple(couplings[i][0]))
                coupling = couplings[i][1]
                result.couplings.append(
                    (
                        pair_tuple,
                        couplings[i][1],
                    )
                )

            # Sort couplings by pairs
            result.couplings.sort(key=lambda x: x[0])

        meta.success = True

        return result, meta

    @override
    def rot(
        self,
        job: JobContext,
        config: RotJobConfig,
    ):
        """
        Perform rotational calculation.

        :param job: JobContext object containing the job information, metadata is stored in job.meta
        :param config: Rotational configuration
        :return: Tuple of (RotResult, MetaData)
        """
        # Set results
        result = RotResult()
        meta = MetaData(job.conf.name)

        jobdir = self._setup(job, "rot")

        # Run sp first
        # NOTE: optical rotation always run in gas-phase (as in old censo)
        self.__prep(job, config, "rot", jobdir, no_solv=True)
        _, spmeta = self._sp(job, config, jobdir=jobdir, prep=False)

        if not spmeta.success:
            meta.success = False
            meta.error = spmeta.error
            return result, meta

        # Set in/out path
        outputpath = os.path.join(jobdir, "escf.out")

        call = [
            str(Path(config.paths.tm) / "escf"),
            "-smpcpus",
            f"{job.omp}",
        ]

        # Run escf
        env = ENVIRON.copy()
        env["PARA_ARCH"] = "SMP"
        env["PARNODES"] = str(job.omp)
        returncode, _ = self._make_call(call, outputpath, jobdir, env=env)

        meta.success = returncode == 0
        if not meta.success:
            meta.error = "unknown_error"
            return result, meta

        # Grab shieldings from the output
        with open(outputpath) as f:
            lines = f.readlines()

        try:
            start = lines.index(next(x for x in lines if "1st frequency" in x))
        except StopIteration:
            meta.success = False
            meta.error = "Could not read specific rotations"
            return result, meta

        lines = lines[start:]

        frequencies = [
            float(line.strip().split()[-1])
            for line in lines
            if "Frequency / nm:" in line
        ]
        velocity_lines = [
            lines[i : i + 15]
            for i in range(len(lines))
            if "rotatory dispersion (velocity" in lines[i]
        ]
        length_lines = [
            lines[i : i + 15]
            for i in range(len(lines))
            if "rotatory dispersion (length" in lines[i]
        ]
        rotations_velocity = [
            float(line.split("(-1)")[-1])
            for linesv in velocity_lines
            for line in linesv
            if "specific" in line
        ]
        rotations_length = [
            float(line.split("(-1)")[-1])
            for linesl in length_lines
            for line in linesl
            if "specific" in line
        ]

        if not len(frequencies) == len(rotations_velocity) == len(rotations_length):
            meta.success = False
            meta.error = "Mismatch in number of frequencies and rotations"
            return result, meta

        for i in range(len(frequencies)):
            result.rotations_velocity.append(
                (
                    frequencies[i],
                    rotations_velocity[i],
                )
            )
            result.rotations_length.append(
                (
                    frequencies[i],
                    rotations_length[i],
                )
            )

        return result, meta

    @override
    def uvvis(self, *args, **kwargs):
        raise NotImplementedError


Factory.register_builder(Prog.TM, TmProc)



#!/usr/bin/env python3

#
# Copyright (C) 2024 Leopold M. Seidler
#
# UVVISPLOT is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# UVVISPLOT is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with UVVISPLOT. If not, see <https://www.gnu.org/licenses/>.

"""
Created on Mar 17, 2024
last updated on 17-March-2024
@author: lmseidler
"""
import matplotlib.pyplot as plt
import os
import argparse
import json
import numpy as np
import pandas as pd

PLANCK = 6.62607015e-34
C = 2.998e8
COULOMB = 1.602e-19


descr = """
     __________________________________________________
    |                                                  |
    |                    UVVISPLOT                     |
    |        Plotting of ensemble UV/Vis spectra       |
    |             University of Bonn, MCTC             |
    |                   March 2024                     |
    |                     v 1.0.0                      |
    |                  L. M. Seidler                   |
    |__________________________________________________|
    """


def get_args():
    """
    Parse command-line arguments.

    :return: Parsed arguments.
    """
    parser = argparse.ArgumentParser(
        description="",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        usage=argparse.SUPPRESS,
    )  # argparse.RawDescriptionHelpFormatter) #,
    parser.add_argument(
        "-mode",
        dest="mode",
        action="store",
        required=False,
        default="wavenumber",
        type=str,
        choices=["wavenumber", "energy", "wavelength"],
        help="Set the unit of the x-axis. Can be wavenumber [cm-1], energy [eV] or wavelength [nm].",
    )
    parser.add_argument(
        "-start",
        dest="start",
        action="store",
        required=False,
        type=float,
        help="Start plotting from '<start>'. Default values: 300nm/1.8eV/14000cm-1.",
    )
    parser.add_argument(
        "-end",
        dest="end",
        action="store",
        required=False,
        type=float,
        help="End plotting at '<end>'. '<end>' must be larger than '<start>'. Default values: 700nm/4.2eV/33000cm-1.",
    )
    parser.add_argument(
        "-title",
        "--title",
        dest="title",
        action="store",
        required=False,
        default="UVVis-PLOT",
        type=str,
        help="Set title of entire plot. If no title is required use " "'<--title ''>'.",
    )
    parser.add_argument(
        "-lw",
        "--linewidth",
        dest="lw",
        action="store",
        required=False,
        default=1.6131e3,
        type=float,
        help="Set linewidth in cm-1.",
    )
    parser.add_argument(
        "-i",
        "--inp",
        dest="inp",
        action="store",
        required=True,
        help="Provide input file.",
    )
    parser.add_argument(
        "-fontsize",
        "--fontsize",
        dest="fontsize",
        action="store",
        required=False,
        default=15,
        type=float,
        help="Set fontsize for entire plot.",
    )
    parser.add_argument(
        "-o",
        "--out",
        dest="out",
        action="store",
        required=False,
        default="nmrplot",
        help="Provide name of the output file (including ending).",
    )
    args = parser.parse_args()
    return args


def read_data(inp):
    """
    Read data from input file.

    :param inp: Input file path.
    :return: Loaded data.
    """
    cwd = os.getcwd()
    with open(os.path.join(cwd, inp)) as f:
        data = json.load(f)

    return data


def plot(data, args):
    """
    Plot the UV-Vis spectrum.

    :param data: The data to plot.
    :param args: Parsed arguments.
    :return: The figure.
    """
    mode = args.mode

    # Select start value
    if args.start is not None:
        start = args.start
    else:
        defaults = {"wavelength": 300, "wavenumber": 14000, "energy": 1.8}
        start = defaults[mode]

    # Select end value
    if args.end is not None:
        end = args.end
    else:
        defaults = {"wavelength": 700, "wavenumber": 33000, "energy": 4.2}
        end = defaults[mode]

    assert end > start
    xrange = np.linspace(start, end, 10000)

    # Dump single contributions to csv file
    confs = {d[2] for d in data}
    exc_number = {conf: 0 for conf in confs}
    contributions_dict = {}

    for exc in data:
        yrange = gaussian_signal(xrange, exc[0], exc[1], args.lw, mode=mode)
        contributions_dict[f"{exc[2]}_S{exc_number[exc[2]]}"] = yrange

        exc_number[exc[2]] += 1

    cwd = os.getcwd()
    contributions_df = pd.DataFrame.from_dict(contributions_dict)
    contributions_df.to_csv(os.path.join(cwd, "contributions.csv"))
    print("All contributions written to contributions.csv.")

    # Plot the whole spectrum
    fig, ax = plt.subplots()
    yrange = contributions_df.sum(axis=1)
    ax.plot(xrange, yrange)
    ax.set_title(args.title)
    labels = {
        "wavelength": r"$\mathrm{nm}$",
        "wavenumber": r"$\mathrm{cm-1}$",
        "energy": r"$\mathrm{eV}$",
    }
    ax.set_xlabel(f"{args.mode} [{labels[args.mode]}]")
    ax.set_ylabel(r"$\epsilon$ [a. u.]")

    return fig


def gaussian_signal(xrange, center_wl, eps_max, lw, mode="wavelength"):
    """
    Generate gaussian signal.

    :param xrange: Range of x values.
    :param center_wl: Center wavelength.
    :param eps_max: Maximum epsilon.
    :param lw: Linewidth.
    :param mode: Mode for units.
    :return: The signal array.
    """
    # <=> 1/λ = E / (h c)
    # 1 nm = 1e-7 cm
    # 1 cm-1 = 1e7 nm-1
    if mode == "wavelength":
        return eps_max * np.exp(-(((1 / xrange - 1 / center_wl) / (lw * 1e7)) ** 2))
    elif mode == "wavenumber":
        return eps_max * np.exp(-(((xrange - 1 / center_wl * 1e7) / lw) ** 2))
    elif mode == "energy":
        return eps_max * np.exp(
            -(
                ((xrange * COULOMB / (PLANCK * C) - 1 / center_wl * 1e9) / (lw * 1e2))
                ** 2
            )
        )


def save_plot(fig, out):
    """
    Save the plot to file.

    :param fig: The figure.
    :param out: Output file path.
    """
    cwd = os.getcwd()
    fig.savefig(os.path.join(cwd, out), format="pdf")


def main():
    """
    Main execution function.
    """

    # Parse cml args
    args = get_args()

    # Read data
    data = read_data(args.inp)

    # Plot data
    figure = plot(data, args)

    # Save plot
    save_plot(figure, args.out)


if __name__ == "__main__":
    main()



"""
Optimization == part2
performing geometry optimization of the CRE and provide low level free energies.
"""

from copy import copy
import json
from pathlib import Path
from collections.abc import Callable
from typing import Any
from tabulate import tabulate
from dask.distributed import Client

from ..ensemble import EnsembleData
from ..molecules import MoleculeData, Contributions
from ..processing import QmProc, XtbProc
from ..parallel import execute
from ..params import AU2KCAL, PLENGTH, GridLevel, Prog
from ..config import PartsConfig
from ..config.parts import OptimizationConfig
from ..config.job_config import RRHOJobConfig, OptJobConfig, XTBOptJobConfig
from ..utilities import (
    printf,
    h1,
    Factory,
    DataDump,
    h2,
)
from ..logging import setup_logger

logger = setup_logger(__name__)


def optimization(
    ensemble: EnsembleData,
    config: PartsConfig,
    client: Client,
    cut: bool = True,
) -> dict[str, Any]:
    """
    Geometry optimization of the ensemble at DFT level (possibly with implicit solvation)

    Uses xtb as driver for orca/tm, calculates hessians from orca/tm single-points and reevaluates ensemble every 'optcycles' steps
    by calculating the thermodynamics of the optimized geometries

    Alternatively just run the complete geometry optimization for every conformer with xtb as driver (decide with 'macrocycles')

    :param ensemble: EnsembleData object containing the conformers.
    :param config: PartsConfig object with configuration settings.
    :param client: dask.distributed.Client for parallel execution.
    :param cut: Whether to apply cutting conditions during optimization.
    :return: JSON-serializable dictionary of optimization results.
    """
    printf(h2("OPTIMIZATION"))

    config = config.model_validate(config, context={"check": "optimization"})

    # Setup processor
    proc = Factory[QmProc].create(config.optimization.prog, "2_OPTIMIZATION")

    if config.optimization.macrocycles:
        contributions_dict = _macrocycle_opt(proc, ensemble, config, client, cut)
    else:
        contributions_dict = _full_opt(proc, ensemble, config, client=client)

    printf("\n")

    # Potentially reevaluate rrho contributions
    if config.general.evaluate_rrho:
        # Run mRRHO calculation
        proc_xtb = Factory[XtbProc].create(Prog.XTB, "2_OPTIMIZATION")
        job_config = RRHOJobConfig(
            gfnv=config.optimization.gfnv,
            paths=config.paths,
            **config.general.model_dump(),
        )
        results = execute(
            ensemble.conformers,
            proc_xtb.xtb_rrho,
            job_config,
            "xtb",
            "optimization",
            client,
            ignore_failed=config.general.ignore_failed,
            balance=config.general.balance,
            copy_mo=config.general.copy_mo,
        )

        if config.general.ignore_failed:
            ensemble.remove_conformers(lambda conf: conf.name not in results)

        for conf in ensemble:
            contributions_dict[conf.name].grrho = results[conf.name].energy

    ensemble.update_contributions(contributions_dict)

    if cut:
        threshold = (
            min(conf.gtot for conf in ensemble)
            + config.optimization.threshold / AU2KCAL
        )
        ensemble.remove_conformers(lambda conf: conf.gtot > threshold)

    # Print/write out results
    results_dict = jsonify(ensemble, config.optimization)
    _write_results(ensemble, config, results_dict)

    return results_dict


def _macrocycle_opt(
    proc: QmProc,
    ensemble: EnsembleData,
    config: PartsConfig,
    client: Client,
    cut: bool,
):
    """
    Geometry optimization using macrocycles, whereafter every macrocycle cutting conditions are checked (if cut == True).

    :param proc: Quantum mechanics processor
    :type proc: QmProc
    :param ensemble: Ensemble data
    :type ensemble: EnsembleData
    :param config: Parts configuration
    :type config: PartsConfig
    :param client: Dask client
    :type client: Client
    :param cut: Whether to apply cutting conditions
    :type cut: bool
    :returns: Contributions dictionary
    :rtype: dict[str, Contributions]
    """
    # Second, independent ensemble containing unconverged conformers
    unconverged_ensemble = EnsembleData()

    # Copy the references to the original conformers into the unconverged_ensemble
    # This way all modifications to the MoleculeData objects will be made on the original objects
    # But the ensembles themselves are independent
    unconverged_ensemble.conformers = copy(ensemble.conformers)

    # Set up contributions_dict
    contributions_dict = {conf.name: Contributions() for conf in ensemble}

    # Set up target
    if config.optimization.xtb_opt:
        opt_job_config = XTBOptJobConfig(
            grid=GridLevel.HIGH,
            copy_mo=config.general.copy_mo,
            gas_phase=config.general.gas_phase,
            solvent=config.general.solvent,
            paths=config.paths,
            **config.optimization.model_dump(),
        )
        if config.optimization.constrain:
            opt_job_config.constraints = ensemble.constraints

        target = proc.xtb_opt
    else:
        opt_job_config = OptJobConfig(
            grid=GridLevel.HIGH,
            copy_mo=config.general.copy_mo,
            gas_phase=config.general.gas_phase,
            solvent=config.general.solvent,
            paths=config.paths,
            **config.optimization.model_dump(),
        )  # type: ignore
        target = proc.opt

    # Iterate and cut only the unconverged conformers
    ncyc = 0
    rrho_done = False
    while (
        len(unconverged_ensemble.conformers) > 0 and ncyc < config.optimization.maxcyc
    ):
        printf(
            h1(
                f"OPTIMIZATION CYCLE {ncyc // config.optimization.optcycles} ({config.optimization.optcycles} steps)"
            )
        )
        results = execute(
            unconverged_ensemble.conformers,
            target,
            opt_job_config,
            config.optimization.prog,
            "optimization",
            client,
            ignore_failed=config.general.ignore_failed,
            balance=config.general.balance,
            copy_mo=config.general.copy_mo,
        )
        if config.general.ignore_failed:
            ensemble.remove_conformers(
                lambda conf: conf in unconverged_ensemble.conformers
                and conf.name not in results
            )
            unconverged_ensemble.remove_conformers(
                lambda conf: conf.name not in results
            )

        for conf in unconverged_ensemble:
            # Update energies
            contributions_dict[conf.name].energy = results[conf.name].energy

            # Update geometries in both ensemble objects
            conf.geom.xyz = results[conf.name].geom

        if (
            config.general.evaluate_rrho
            and ncyc + config.optimization.optcycles > 6
            and not rrho_done
        ):
            # Run mRRHO calculation
            proc_xtb: XtbProc = Factory.create(Prog.XTB, "2_OPTIMIZATION")
            job_config_rrho = RRHOJobConfig(
                gfnv=config.optimization.gfnv,
                paths=config.paths,
                **config.general.model_dump(),
            )
            results_rrho = execute(
                unconverged_ensemble.conformers,
                proc_xtb.xtb_rrho,
                job_config_rrho,
                "xtb",
                "optimization",
                client,
                ignore_failed=config.general.ignore_failed,
                balance=config.general.balance,
                copy_mo=config.general.copy_mo,
            )
            if config.general.ignore_failed:
                ensemble.remove_conformers(
                    lambda conf: conf in unconverged_ensemble.conformers
                    and conf.name not in results_rrho
                )
                unconverged_ensemble.remove_conformers(
                    lambda conf: conf.name not in results_rrho
                )

            for conf in unconverged_ensemble:
                contributions_dict[conf.name].grrho = results_rrho[conf.name].energy

            rrho_done = True

        # TODO: molbar topology check

        # Remove converged conformers
        unconverged_ensemble.remove_conformers(
            lambda conf: results[conf.name].converged
        )

        # Update contributions
        unconverged_ensemble.update_contributions(contributions_dict)

        if len(unconverged_ensemble.conformers) != 0 and cut:
            gradthr = config.optimization.gradthr

            # Take the minimum Gtot for all the conformers (converged AND unconverged) (converged are considered removed from the unconverged_ensemble)
            threshold = (
                min(
                    [conf.gtot for conf in unconverged_ensemble]
                    + [conf.gtot for conf in unconverged_ensemble.rem]
                )
                + config.optimization.threshold / AU2KCAL
            )
            unconverged_ensemble.remove_conformers(
                lambda conf: conf.gtot > threshold
                and results[conf.name].grad_norm < gradthr
            )
        # if all confs are converged, don't cut again, instead let the cutting be done in the optimization function,
        # so we can use the updated mRRHO contributions
        ncyc += config.optimization.optcycles

        _print_update(unconverged_ensemble)

    for conf in unconverged_ensemble:
        logger.warning(
            f"{conf.name} did not converge after {config.optimization.maxcyc} cycles."
        )

    ensemble.remove_conformers(lambda conf: conf in unconverged_ensemble)

    # Return most up to date information
    return contributions_dict


def _full_opt(
    proc: QmProc,
    ensemble: EnsembleData,
    config: PartsConfig,
    client: Client,
):
    """
    Full geometry optimization of every conformer.

    :param proc: Quantum mechanics processor
    :type proc: QmProc
    :param ensemble: Ensemble data
    :type ensemble: EnsembleData
    :param config: Parts configuration
    :type config: PartsConfig
    :param client: Dask client
    :type client: Client
    :returns: Contributions dictionary
    :rtype: dict[str, Contributions]
    """
    # Set up target
    if config.optimization.xtb_opt:
        opt_job_config = XTBOptJobConfig(
            copy_mo=config.general.copy_mo,
            grid=GridLevel.HIGH,
            gas_phase=config.general.gas_phase,
            solvent=config.general.solvent,
            paths=config.paths,
            **config.optimization.model_dump(),
        )
        if config.optimization.constrain:
            opt_job_config.constraints = ensemble.constraints

        target = proc.xtb_opt
    else:
        opt_job_config = OptJobConfig(
            copy_mo=config.general.copy_mo,
            grid=GridLevel.HIGH,
            gas_phase=config.general.gas_phase,
            solvent=config.general.solvent,
            paths=config.paths,
            **config.optimization.model_dump(),
        )  # type: ignore
        target = proc.opt

    # Set up contributions_dict
    contributions_dict = {conf.name: Contributions() for conf in ensemble}

    results = execute(
        ensemble.conformers,
        target,
        opt_job_config,
        config.optimization.prog,
        "optimization",
        client,
        ignore_failed=config.general.ignore_failed,
        balance=config.general.balance,
        copy_mo=config.general.copy_mo,
    )

    if config.general.ignore_failed:
        ensemble.remove_conformers(lambda conf: conf.name not in results)

    for conf in ensemble:
        contributions_dict[conf.name].energy = results[conf.name].energy
        conf.geom.xyz = results[conf.name].geom

    return contributions_dict


def _write_results(
    ensemble: EnsembleData, config: PartsConfig, results_dict: dict[str, Any]
) -> None:
    """ """
    printf(h1("OPTIMIZATION RESULTS"))

    # column headers
    headers = [
        "CONF#",
        # "G (xTB)",
        # "ΔG (xTB)",
        "E (DFT)",
        "ΔGsolv",
        "GmRRHO",
        "Gtot",
        "ΔGtot",
        "Boltzmann weight",
    ]

    # column units
    units = [
        "",
        # "[Eh]",
        # "[kcal/mol]",
        "[Eh]",
        "[kcal/mol]",
        "[Eh]",
        "[Eh]",
        "[kcal/mol]",
        f"% at {config.general.temperature} K",
    ]

    # variables for printmap
    gtotmin = min(conf.gtot for conf in ensemble)

    boltzmann_populations = ensemble.get_populations(config.general.temperature)

    printmap = {
        "CONF#": lambda conf: conf.name,
        # "G (xTB)": lambda conf: (
        #     f"{gxtb[conf.name]:.6f}" if gxtb is not None else "---"
        # ),
        # "ΔG (xTB)": lambda conf: (
        #     f"{(gxtb[conf.name] - gxtbmin) * AU2KCAL:.2f}"
        #     if gxtb is not None and gxtbmin is not None
        #     else "---"
        # ),
        "E (DFT)": lambda conf: f"{conf.energy:.6f}",
        "ΔGsolv": lambda conf: (f"{conf.gsolv * AU2KCAL:.6f}"),
        "GmRRHO": lambda conf: (f"{conf.grrho:.6f}"),
        "Gtot": lambda conf: f"{conf.gtot:.6f}",
        "ΔGtot": lambda conf: f"{(conf.gtot - gtotmin) * AU2KCAL:.2f}",
        "Boltzmann weight": lambda conf: f"{boltzmann_populations[conf.name] * 100:.2f}",
    }

    rows = [[printmap[header](conf) for header in headers] for conf in ensemble]

    for i in range(len(headers)):
        headers[i] += "\n" + units[i]

    table = tabulate(
        rows,
        headers=headers,
        colalign=["left"] + ["center" for _ in headers[1:]],
        disable_numparse=True,
        numalign="decimal",
    )
    printf(table)

    # list the averaged free enthalpy of the ensemble
    lines: list[str] = []
    lines.append(
        "\nBoltzmann averaged free energy/enthalpy of ensemble on input geometries (DFT optimized):"
    )
    lines.append(f"{'temperature /K:':<15} {'avE(T) /a.u.':>14} {'avG(T) /a.u.':>14}")

    # calculate averaged free enthalpy
    avG = sum([boltzmann_populations[conf.name] * conf.gtot for conf in ensemble])

    # calculate averaged free energy (?)
    avE = sum([boltzmann_populations[conf.name] * conf.energy for conf in ensemble])

    # append the lines for the free energy/enthalpy
    lines.append(
        f"{config.general.temperature:^15} {avE:>14.7f}  {avG:>14.7f}     <<==part2=="
    )
    lines.append("".ljust(int(PLENGTH), "-"))

    # lines.append(f">>> END of {self.__class__.__name__} <<<".center(PLENGTH, " ") + "\n")

    # Print everything
    for line in lines:
        printf(line)

    # write everything to a file
    filepath = Path("2_OPTIMIZATION.out")
    filepath.write_text(table + "\n".join(lines))

    # Additionally, write results in json format
    Path("2_OPTIMIZATION.json").write_text(json.dumps(results_dict, indent=4))

    ensemble.dump_xyz(Path("2_OPTIMIZATION.xyz"))


# TODO: generalize this
def jsonify(
    ensemble: EnsembleData,
    config: OptimizationConfig,
    fields: Callable[[MoleculeData], dict[str, Any]] | None = None,
):
    """
    Convert ensemble data to JSON format for optimization results.

    :param ensemble: EnsembleData object.
    :param config: OptimizationConfig object.
    :param fields: Optional callable to customize fields.
    :return: JSON-serializable dictionary.
    """
    per_conf: Callable[[MoleculeData], dict[str, dict[str, float]]] = fields or (
        lambda conf: {
            conf.name: {
                "energy": conf.energy,
                "gsolv": conf.gsolv,
                "grrho": conf.grrho,
                "gtot": conf.gtot,
            }
        }
    )

    dump = DataDump(part_name="optimization")

    for conf in ensemble:
        dump.data.update(per_conf(conf))

    dump.settings = config.model_dump()

    return dump.model_dump(mode="json")


def _print_update(ensemble: EnsembleData):
    # column headers
    headers = [
        "CONF#",
        "E (DFT)",
        "GmRRHO",
        "Gtot",
        "ΔGtot",
    ]

    # column units
    units = [
        "",
        "[Eh]",
        "[Eh]",
        "[Eh]",
        "[kcal/mol]",
    ]

    # variables for printmap
    gtotmin = min(conf.gtot for conf in ensemble.conformers + ensemble.rem)

    printmap = {
        "CONF#": lambda conf: conf.name,
        "E (DFT)": lambda conf: f"{conf.energy:.6f}",
        "GmRRHO": lambda conf: (f"{conf.grrho:.6f}"),
        "Gtot": lambda conf: f"{conf.gtot:.6f}",
        "ΔGtot": lambda conf: f"{(conf.gtot - gtotmin) * AU2KCAL:.2f}",
    }

    if len(ensemble.rem) > 0:
        printf(h1("Converged or removed conformers"))

        rows = [[printmap[header](conf) for header in headers] for conf in ensemble.rem]

        for i in range(len(headers)):
            headers[i] += "\n" + units[i]

        table = tabulate(
            rows,
            headers=headers,
            colalign=["left"] + ["center" for _ in headers[1:]],
            disable_numparse=True,
            numalign="decimal",
        )
        printf(table)

        for i in range(len(headers)):
            headers[i] = headers[i].split("\n")[0]

    if len(ensemble.conformers) > 0:
        printf(h1("Unconverged conformers"))

        rows = [[printmap[header](conf) for header in headers] for conf in ensemble]

        for i in range(len(headers)):
            headers[i] += "\n" + units[i]

        table = tabulate(
            rows,
            headers=headers,
            colalign=["left"] + ["center" for _ in headers[1:]],
            disable_numparse=True,
            numalign="decimal",
        )
        printf(table)



"""
Processor for xtb calculations.
"""

import json
import os
from pathlib import Path
from typing import final, override


from ..config.job_config import (
    RRHOJobConfig,
    XTBJobConfig,
)
from .job import (
    JobContext,
)
from .results import (
    GsolvResult,
    MetaData,
    RRHOResult,
    SPResult,
)
from ..utilities import Factory
from ..logging import setup_logger
from ..assets import SOLVENTS
from .processor import GenericProc
from ..params import Prog

logger = setup_logger(__name__)


@final
class XtbProc(GenericProc):
    """
    XtbProcessor
    """

    progname = Prog.XTB

    def _sp(
        self,
        job: JobContext,
        config: XTBJobConfig,
        jobdir: str | Path | None = None,
        filename: str = "xtb_sp",
        no_solv: bool = False,
    ) -> tuple[SPResult, MetaData]:
        """
        Calculates the single-point energy with GFNn-xTB or GFN-FF.
        Unwrapped function to call from other methods.

        :param job: job to run
        :type job: JobContext
        :param config: XTB configuration
        :type config: XTBJobConfig
        :param jobdir: path to the jobdir
        :type jobdir: str | Path | None
        :param filename: filename to use for the coord file. Defaults to "xtb_sp".
        :type filename: str
        :param no_solv: whether to run the sp in gas-phase. Defaults to False.
        :type no_solv: bool
        :returns: Tuple of (SPResult, MetaData)
        :rtype: tuple[SPResult, MetaData]
        """
        if jobdir is None:
            jobdir = self._setup(job, "xtb_sp")

        result = SPResult()
        meta = MetaData(job.conf.name)

        # set in/out path
        inputpath = os.path.join(jobdir, f"{filename}.coord")
        outputpath = os.path.join(jobdir, f"{filename}.out")
        xcontrolname = "xtb_sp-xcontrol-inp"

        # cleanup
        files = [
            "xtbrestart",
            "xtbtopo.mol",
            xcontrolname,
            "wbo",
            "charges",
            "gfnff_topo",
            f"{filename}.out",
        ]

        # remove potentially preexisting files to avoid confusion
        for file in files:
            if os.path.isfile(os.path.join(jobdir, file)):
                os.remove(os.path.join(jobdir, file))

        # generate coord file for xtb
        with open(inputpath, "w", newline=None) as f:
            f.writelines(job.conf.tocoord())

        # setup call for xtb single-point
        call: list[str] = [
            config.paths.xtb,
            f"{filename}.coord",
            "--" + config.gfnv,
            "--sp",
            "--chrg",
            f"{job.charge}",
            "--norestart",
            "--parallel",
            f"{job.omp}",
        ]

        # add solvent to xtb call if not a gas-phase sp
        if not no_solv and not config.gas_phase:
            assert config.solvent
            assert config.sm_rrho
            solvent_key = SOLVENTS[config.solvent][config.sm_rrho]
            call.extend(
                [
                    "--" + config.sm_rrho,
                    solvent_key,
                    "reference",
                    "-I",
                    xcontrolname,
                ]
            )

            # set gbsa grid
            with open(os.path.join(jobdir, xcontrolname), "w", newline=None) as xcout:
                xcout.writelines(["$gbsa\n", "  gbsagrid=tight\n", "$end\n"])

        # call xtb
        returncode, _ = self._make_call(call, outputpath, jobdir)

        # if returncode != 0 then some error happened in xtb
        # TODO: returncodes
        if returncode != 0:
            meta.success = False
            meta.error = "unknown_error"
            return result, meta

        # read energy from outputfile
        with open(outputpath) as outputfile:
            for line in outputfile.readlines():
                if "| TOTAL ENERGY" in line:
                    result.energy = float(line.split()[3])
                    meta.success = True
                    break
                if "convergence criteria cannot be satisfied" in line:
                    meta.success = False
                    meta.error = "scf_not_converged"
                    break

        return result, meta

    @override
    def sp(self, *args, **kwargs):
        """
        Wrapped version of _sp.

        :param args: Arguments.
        :param kwargs: Keyword arguments.
        :return: Tuple of (SPResult, MetaData).
        """
        return self._sp(*args, **kwargs)

    def gsolv(
        self, job: JobContext, config: XTBJobConfig
    ) -> tuple[GsolvResult, MetaData]:
        """
        Calculate additive GBSA or ALPB solvation using GFNn-xTB or GFN-FF.

        :param job: JobContext object containing the job information, metadata is stored in job.meta
        :type job: JobContext
        :param config: XTB configuration
        :type config: XTBJobConfig
        :returns: Tuple of (GsolvResult, MetaData)
        :rtype: tuple[GsolvResult, MetaData]
        """
        result = GsolvResult()
        meta = MetaData(job.conf.name)

        jobdir = self._setup(job, "xtb_gsolv")

        # run gas-phase GFN single-point
        spres, spmeta = self._sp(
            job, config, jobdir=jobdir, filename="gas", no_solv=True
        )
        if spmeta.success:
            result.energy_gas = spres.energy
        else:
            meta.success = False
            meta.error = spmeta.error
            return result, meta

        # run single-point in solution:
        spres, spmeta = self._sp(job, config, jobdir=jobdir, filename="solv")
        if spmeta.success:
            result.energy_solv = spres.energy
        else:
            meta.success = False
            meta.error = spmeta.error
            return result, meta

        # only reached if both gas-phase and solvated sp succeeded
        result.gsolv = result.energy_solv - result.energy_gas
        meta.success = True

        return result, meta

    # TODO: break this down
    def xtb_rrho(
        self,
        job: JobContext,
        config: RRHOJobConfig,
    ) -> tuple[RRHOResult, MetaData]:
        """
        Calculates the mRRHO contribution to the free enthalpy of a conformer with GFNn-xTB/GFN-FF.

        :param job: job to run
        :type job: JobContext
        :param config: RRHO configuration
        :type config: RRHOJobConfig
        :returns: result of the rrho calculation and metadata
        :rtype: tuple[RRHOResult, MetaData]
        """
        # what is returned in the end
        result = RRHOResult()
        meta = MetaData(job.conf.name)

        jobdir = self._setup(job, "xtb_rrho")
        filename = "xtb_rrho"

        # set in/out path
        outputpath = os.path.join(jobdir, f"{filename}.out")
        xcontrolname = "rrho-xcontrol-inp"
        xcontrolpath = os.path.join(jobdir, xcontrolname)

        # TODO - is this list complete?
        files = [
            "xtbrestart",
            "xtbtopo.mol",
            xcontrolname,
            "wbo",
            "charges",
            "gfnff_topo",
        ]

        # remove potentially preexisting files to avoid confusion
        for file in files:
            if os.path.isfile(os.path.join(jobdir, file)):
                os.remove(os.path.join(jobdir, file))

        # setup xcontrol
        with open(xcontrolpath, "w", newline=None) as xcout:
            xcout.write("$thermo\n")
            # if config.multitemp:
            #     trange = frange(
            #         config.trange[0],
            #         config.trange[1],
            #         step=config.trange[2],
            #     )
            #
            #     # Always append the fixed temperature to the trange so that it is the last value
            #     # (important since --enso will make xtb give the G(T) value for this temperature)
            #     assert config.temperature
            #     trange.append(config.temperature)
            #
            #     # Write trange to the xcontrol file
            #     xcout.write(f"    temp=" + ",".join([str(i) for i in trange]) + "\n")
            #     if config.temperature in trange[:-1]:
            #         trange = trange[:-1]
            # else:
            xcout.write(f"    temp={config.temperature}\n")

            xcout.write(f"    sthr={config.sthr}\n")

            xcout.write(f"    imagthr={config.imagthr}\n")

            # TODO - when do you actually want to set the scale manually?
            # don't set scale manually for now
            """ if self.instructions.get("scale", "automatic") != "automatic":
                # for automatic --> is method dependant leave it to xTB e.g. GFNFF has a
                # different scaling factor than GFN2
                xcout.write(f"    scale={self.instructions['scale']}\n") """

            xcout.write("$symmetry\n")
            xcout.write("     maxat=1000\n")
            # always consider symmetry
            # xcout.write("    desy=0.1\n") # taken from xtb defaults
            # xcout.write("    desy=0.0\n")

            # set gbsa grid
            if not config.gas_phase:
                xcout.write("$gbsa\n")
                xcout.write("  gbsagrid=tight\n")

            xcout.write("$end\n")

        # if config.bhess:
        #     # set ohess or bhess
        #     dohess = "--bhess"
        # else:
        #     dohess = "--ohess"

        # generate coord file for xtb
        with open(os.path.join(jobdir, f"{filename}.coord"), "w", newline=None) as f:
            f.writelines(job.conf.tocoord())

        call: list[str] = [
            config.paths.xtb,
            f"{filename}.coord",
            "--" + config.gfnv,
            "--bhess",
            "vtight",
            "--chrg",
            f"{job.charge}",
            "--enso",
            "--norestart",
            "-I",
            xcontrolname,
            "--parallel",
            f"{job.omp}",
        ]

        # add solvent to xtb call if necessary
        if not config.gas_phase:
            assert config.solvent
            assert config.sm_rrho
            solvent_key = SOLVENTS[config.solvent][config.sm_rrho]
            call.extend(
                [
                    "--" + config.sm_rrho,
                    solvent_key,
                ]
            )

        # TODO: if rmsd bias is used (should be defined in censo workdir (cwd))
        # if config.rmsdbias:
        #     cwd = os.getcwd()
        #
        #     call.extend(
        #         [
        #             "--bias-input",
        #             os.path.join(cwd, "rmsdpot.xyz"),
        #         ]
        #     )

        # call xtb
        returncode, errors = self._make_call(call, outputpath, jobdir)

        # check if converged:
        if returncode != 0:
            meta.success = False
            meta.error = "unknown_error"
            return result, meta

        # read output and store lines
        with open(outputpath) as outputfile:
            lines = outputfile.readlines()

        # if config.multitemp:
        #     # get gibbs energy, enthalpy and entropy for given temperature range
        #     # gibbs energy
        #     gt: dict[float, float] = {}
        #
        #     # enthalpy
        #     ht: dict[float, float] = {}
        #
        #     # Get Gibbs energy and enthalpy
        #     for i, line in enumerate(lines):
        #         if "T/K" in line:
        #             for line2 in lines[i + 2 :]:
        #                 if "----------------------------------" in line2:
        #                     break
        #
        #                 temp = float(line2.split()[0])
        #                 gt[temp] = float(line2.split()[4])
        #                 ht[temp] = float(line2.split()[2])
        #             break
        #
        #     # rotational entropy
        #     rotS: dict[float, float] = {}
        #
        #     # Get rotational entropy
        #     entropy_lines = [
        #         (line, lines[i + 1]) for i, line in enumerate(lines) if "VIB" in line
        #     ]
        #     for line in entropy_lines:
        #         temp = float(line[0].split()[0])
        #         rotS[temp] = float(line[1].split()[4])
        #
        #     # check if xtb calculated the temperature range correctly
        #     if not (
        #         len(trange) == len(gt)
        #         and len(trange) == len(ht)
        #         and len(trange) == len(rotS)
        #     ):
        #         meta.success = False
        #         meta.error = "xtb_trange_inconsistent"
        #         return result, meta
        #     else:
        #         result.gibbs = gt
        #         result.enthalpy = ht
        #         result.entropy = rotS

        # Extract symmetry
        result.linear = next(
            {"true": True, "false": False}[line.split()[2]]
            for line in lines
            if ":  linear? " in line
        )

        # Extract rmsd
        # if config.bhess:
        #     result.rmsd = next(
        #         (float(line.split()[3]) for line in lines if "final rmsd / " in line)
        #     )

        # xtb_enso.json is generated by xtb by using the '--enso' argument *only* when using --bhess or --ohess
        # (when a hessian is calculated)
        # contains output from xtb in json format to be more easily digestible by CENSO
        with open(
            os.path.join(jobdir, "xtb_enso.json"),
        ) as f:
            data = json.load(f)

        # read number of imaginary frequencies and print warning
        if "number of imags" in data:
            if data["number of imags"] > 0:
                logger.warning(
                    f"Found {data['number of imags']} significant"
                    + " imaginary frequencies for "
                    + f"{job.conf.name}."
                )

        # get gibbs energy
        if "G(T)" in data:
            meta.success = True

            temp = float(config.temperature)
            if config.temperature == 0.0:
                result.energy = data.get("ZPVE", 0.0)
                result.gibbs[temp] = data.get("ZPVE", 0.0)
                result.enthalpy[temp] = data.get("ZPVE", 0.0)
                # if not config.multitemp:
                result.entropy[temp] = 0.0  # set this to None for predictability (?)
            else:
                result.energy = data.get("G(T)", 0.0)
                result.gibbs[temp] = data.get("G(T)", 0.0)
                # if not config.multitemp:
                result.enthalpy[temp] = 0.0  # set this to None for predictability (?)
                result.entropy[temp] = 0.0  # set this to None for predictability (?)

            # only determine symmetry if all the needed information is there
            if "point group" and "linear" in data:
                result.symmetry = data["point group"]
                result.linear = data["linear"]
            else:
                # could not determine symmetry correctly
                result.symmetry = "c1"
                result.linear = False

            # calculate symnum
            result.symnum = self._get_sym_num(sym=result.symmetry, linear=result.linear)
        else:
            meta.success = False
            meta.error = "Could not read xtb_enso.json"

        return result, meta

    @override
    def opt(self, *args, **kwargs):
        raise NotImplementedError


Factory.register_builder(Prog.XTB, XtbProc)



from typing import override, Self
import enum
import warnings

from pydantic import model_validator


from ...assets import FUNCTIONALS
from ...params import PLENGTH, QmProg
from ..generic import GenericConfig
from ...utilities import h2


class BasePartConfig(GenericConfig):
    """Base class for configuration classes for all parts"""

    @override
    def __str__(self) -> str:
        lines: list[str] = []
        lines.append(h2(f"{self.__class__.__name__.split("Config")[0]}"))
        kv = [(k, v) for k, v in self]
        kv.sort(key=lambda x: type(x[1]).__name__)
        for name, value in kv:
            display_value = value.value if isinstance(value, enum.Enum) else value
            lines.append(f"{name:>{PLENGTH // 2 - 2}} : {display_value}")

        return str("\n".join(lines))

    @model_validator(mode="after")
    def tm_gcp_d4_bug_check(self) -> Self:
        """
        Check if the model has basis, func, and prog fields and get their values. Backcheck with GCP basis sets.

        :return: The validated instance.
        """
        if all(s in self.__class__.model_fields for s in ["func", "basis", "prog"]):
            func: str = getattr(self, "func")

            try:
                disp = FUNCTIONALS[func]["disp"]
            except KeyError:
                raise ValueError("Received invalid functional key.")

            gcp_keywords = {
                "minis": "MINIS",
                "sv": "SV",
                "6-31g(d)": "631GD",
                "def2-sv(p)": "SV(P)",
                "def2-svp": "SVP",
                "def2-tzvp": "TZ",
            }
            if (
                getattr(self, "basis") in gcp_keywords
                and disp == "d4"
                and getattr(self, "prog") == QmProg.TM.value
            ):
                warnings.warn(
                    "Small basis set detected: Current version of TM includes a bug when combining GCP with DFT-D4. Switching to D3."
                )
                setattr(self, "func", func.replace("d4", "d3"))
                # TODO: there should be a way to validate this afterwards

        return self



import time
import pytest
import subprocess
from typing import Any
from math import ceil

from censo.config.paths import PathsConfig
from censo.parallel import (
    set_omp,
    prepare_jobs,
    execute,
)
from censo.processing.job import JobContext
from censo.processing.results import (
    SPResult,
    GsolvResult,
    RRHOResult,
    OptResult,
    NMRResult,
    UVVisResult,
    MetaData,
)
from censo.params import GridLevel, QmProg
from censo.config.job_config import SPJobConfig


# ============= Fixtures and Mock Classes =============


class MockGeometryData:
    """Mock class for GeometryData"""

    def __init__(self, name: str):
        self.name = name
        self.xyz = ["C 0.0 0.0 0.0\n", "H 1.0 0.0 0.0\n"]


@pytest.fixture
def molecule_data():
    """Create a real MoleculeData instance"""
    from censo.molecules import MoleculeData

    mol = MoleculeData(
        name="CONF1", xyz=["C 0.0 0.0 0.0\n", "H 1.0 0.0 0.0\n"], charge=0, unpaired=0
    )
    mol.mo_paths = {
        QmProg.TM: ["path/to/mos"],
        QmProg.ORCA: ["path/to/file.gbw"],
    }
    return mol


@pytest.fixture
def parallel_job(molecule_data):
    """Create a real JobContext instance"""
    return JobContext(
        conf=molecule_data.geom,
        charge=molecule_data.charge,
        unpaired=molecule_data.unpaired,
        omp=1,
    )


@pytest.fixture
def mock_job_config():
    """Create a mock job configuration"""
    return SPJobConfig(
        copy_mo=False,
        func="pbe-d4",
        basis="def2-sv(p)",
        grid=GridLevel.LOW,
        template=False,
        gas_phase=True,
        paths=PathsConfig.model_construct(**{}),
    )


@pytest.fixture
def create_conformers():
    """Factory fixture to create specified number of real conformers"""

    def _create_conformers(count: int):
        from censo.molecules import MoleculeData

        conformers = []
        for i in range(1, count + 1):
            mol = MoleculeData(
                name=f"CONF{i}",
                xyz=["C 0.0 0.0 0.0\n", "H 1.0 0.0 0.0\n"],
                charge=0,
                unpaired=0,
            )
            mol.mo_paths = {
                QmProg.TM: ["path/to/mos"],
                QmProg.ORCA: ["path/to/file.gbw"],
            }
            conformers.append(mol)
        return conformers

    return _create_conformers


# ============= Mock Task Functions =============


# Define task functions at module level for pickling
def task_success(
    job: JobContext, job_config: Any, **kwargs
) -> tuple[SPResult, MetaData]:
    """Task that always succeeds"""
    time.sleep(0.1)  # Shorter sleep for faster tests
    return SPResult(mo_path=f"path/to/mo_{job.conf.name}", energy=-100.0), MetaData(
        conf_name=job.conf.name, success=True
    )


def task_fail(job: JobContext, job_config: Any, **kwargs) -> tuple[SPResult, MetaData]:
    """Task that always fails"""
    time.sleep(0.1)
    return SPResult(mo_path="", energy=0.0), MetaData(
        conf_name=job.conf.name, success=False, error="Test error"
    )


def task_conditional(
    job: JobContext, job_config: Any, **kwargs
) -> tuple[Any, MetaData]:
    """Task that succeeds for even-numbered conformers and fails for odd-numbered ones"""
    conf_num = int(job.conf.name.replace("CONF", ""))
    if conf_num % 2 == 0:
        return task_success(job, job_config, **kwargs)
    return task_fail(job, job_config, **kwargs)


# Module-level task for pickling
def task_raise_then_long(job: JobContext, job_config: Any, **kwargs):
    """Raise for CONF1 to trigger cancellation; other jobs sleep longer then succeed."""
    if job.conf.name == "CONF1":
        time.sleep(0.01)
        raise RuntimeError("Intentional failure for cancellation test")
    # Long-running work to give cancellation a chance
    time.sleep(0.1)
    return SPResult(mo_path=f"path/to/mo_{job.conf.name}", energy=-100.0), MetaData(
        conf_name=job.conf.name, success=True
    )


# Module-level task for subprocess cancellation test
def task_long_subprocess(
    job: JobContext, job_config: Any, **kwargs
) -> tuple[SPResult, MetaData]:
    """Task that runs a long subprocess and checks for cancellation."""
    from dask.distributed import Variable, get_client

    # Get cancel var name from kwargs
    cancel_var_name = kwargs.get("__censo_cancel_var")
    if cancel_var_name:
        try:
            client = get_client()
        except Exception:
            client = None
        cancel_var = Variable(cancel_var_name, client=client)
    else:
        cancel_var = None

    # Run a long sleep command
    proc = subprocess.Popen(
        ["sleep", "10"], stdout=subprocess.PIPE, stderr=subprocess.PIPE
    )

    # Poll for completion or cancellation
    while proc.poll() is None:
        if cancel_var and cancel_var.get():
            proc.terminate()
            try:
                proc.wait(timeout=1)
            except subprocess.TimeoutExpired:
                proc.kill()
            return SPResult(mo_path="", energy=0.0), MetaData(
                conf_name=job.conf.name, success=False, error="cancelled"
            )
        time.sleep(0.1)

    # If not cancelled, succeed
    return SPResult(mo_path=f"path/to/mo_{job.conf.name}", energy=-100.0), MetaData(
        conf_name=job.conf.name, success=True
    )


# ============= Tests for Core Functionality =============


class TestResultDataclasses:
    """Tests for QM result dataclasses"""

    def test_sp_result(self):
        """Test SPResult dataclass"""
        sp_result = SPResult(mo_path="path/to/mo", energy=-100.0)
        assert sp_result.mo_path == "path/to/mo"
        assert sp_result.energy == -100.0

    def test_gsolv_result(self):
        """Test GsolvResult dataclass"""
        gsolv_result = GsolvResult(
            mo_path="path/to/mo", gsolv=-0.1, energy_gas=-100.0, energy_solv=-100.1
        )
        assert gsolv_result.mo_path == "path/to/mo"
        assert gsolv_result.gsolv == -0.1
        assert gsolv_result.energy_gas == -100.0
        assert gsolv_result.energy_solv == -100.1

    def test_rrho_result(self):
        """Test RRHOResult dataclass"""
        rrho_result = RRHOResult(
            mo_path="path/to/mo",
            energy=-100.0,
            rmsd=0.001,
            gibbs={298.15: -99.9},
            symmetry="c2v",
        )
        assert rrho_result.mo_path == "path/to/mo"
        assert rrho_result.energy == -100.0
        assert rrho_result.rmsd == 0.001
        assert rrho_result.gibbs[298.15] == -99.9
        assert rrho_result.symmetry == "c2v"

    def test_opt_result(self):
        """Test OptResult dataclass"""
        opt_result = OptResult(
            mo_path="path/to/mo",
            energy=-100.0,
            ecyc=[-99.0, -99.5, -100.0],
            converged=True,
        )
        assert opt_result.mo_path == "path/to/mo"
        assert opt_result.energy == -100.0
        assert len(opt_result.ecyc) == 3
        assert opt_result.converged

    def test_nmr_result(self):
        """Test NMRResult dataclass"""
        nmr_result = NMRResult(
            mo_path="path/to/mo", shieldings=[(1, 30.5)], couplings=[((1, 2), 7.2)]
        )
        assert nmr_result.mo_path == "path/to/mo"
        assert len(nmr_result.shieldings) == 1
        assert len(nmr_result.couplings) == 1
        assert nmr_result.shieldings[0] == (1, 30.5)
        assert nmr_result.couplings[0] == ((1, 2), 7.2)

    def test_uvvis_result(self):
        """Test UVVisResult dataclass"""
        uvvis_result = UVVisResult(
            mo_path="path/to/mo", excitations=[{"wavelength": 350.0, "f": 0.5}]
        )
        assert uvvis_result.mo_path == "path/to/mo"
        assert len(uvvis_result.excitations) == 1
        assert uvvis_result.excitations[0]["wavelength"] == 350.0
        assert uvvis_result.excitations[0]["f"] == 0.5


class TestMetaData:
    """Tests for MetaData class"""

    def test_metadata_success(self):
        """Test MetaData with successful execution"""
        meta = MetaData(conf_name="CONF1", success=True)
        assert meta.conf_name == "CONF1"
        assert meta.success
        assert meta.error == ""

    def test_metadata_failure(self):
        """Test MetaData with failed execution"""
        meta_failed = MetaData(conf_name="CONF2", success=False, error="Test error")
        assert meta_failed.conf_name == "CONF2"
        assert not meta_failed.success
        assert meta_failed.error == "Test error"


class TestJobContext:
    """Tests for JobContext class"""

    def test_parallel_job_initialization(self, molecule_data):
        """Test JobContext initialization and properties"""
        job = JobContext(molecule_data.geom, 0, 0, 1)
        assert job.conf == molecule_data.geom
        assert job.omp == 1
        assert job.charge == 0
        assert job.unpaired == 0
        assert isinstance(job.flags, dict)


# ============= Tests for OMP Threading Configuration =============


class TestOMPConfiguration:
    """Tests for OMP threading configuration functions"""

    OMPMIN = 1  # Local constant for ompmin

    @pytest.mark.parametrize(
        "balance,omp,ncores,njobs,expected_omp",
        [
            (True, 2, 4, 2, 2),  # Test with balancing enabled
            (False, 3, 4, 2, 3),  # Test with balancing disabled
            (True, 2, 8, 2, 4),  # Test optimal distribution
        ],
    )
    def test_set_omp(self, balance, omp, ncores, njobs, expected_omp, parallel_job):
        """Test set_omp function distributes threads properly"""
        jobs = [parallel_job for _ in range(njobs)]
        set_omp(jobs, balance, omp, ncores)
        assert all(job.omp == expected_omp for job in jobs)


# ============= Tests for Job Preparation =============


class TestJobPreparation:
    """Tests for job preparation functions"""

    @pytest.mark.parametrize(
        "prog,copy_mo,expected_mo_guess",
        [
            (QmProg.ORCA, True, "path/to/file.gbw"),
            (QmProg.TM, True, "path/to/mos"),
            (QmProg.ORCA, False, None),
        ],
    )
    def test_prepare_jobs(self, prog, copy_mo, expected_mo_guess, molecule_data):
        """Test prepare_jobs function creates jobs with correct parameters"""
        conformers = [molecule_data]
        jobs = prepare_jobs(
            conformers, prog, 4, 2, "test", balance=True, copy_mo=copy_mo
        )

        assert len(jobs) == 1
        assert isinstance(jobs[0], JobContext)
        if copy_mo:
            assert str(jobs[0].mo_guess) == expected_mo_guess
        else:
            assert jobs[0].mo_guess is None


# ============= Tests for Job Execution =============


class TestJobExecution:
    """Tests for job execution functions"""

    def test_execute_success(self, mock_job_config, create_conformers, parallel_setup):
        """Test successful parallel execution of all jobs"""
        # Create two mock conformers
        conformers = create_conformers(2)
        client, cluster = parallel_setup

        # Run execute
        results = execute(
            conformers=conformers,
            task=task_success,  # type: ignore[arg-type]
            job_config=mock_job_config,
            prog=QmProg.ORCA,
            from_part="test",
            client=client,
            ignore_failed=True,
            balance=True,
            copy_mo=True,
        )

        # Verify results
        assert len(results) == 2
        assert all(f"CONF{i}" in results for i in range(1, 3))

        # Check result contents
        for conf_name, result in results.items():
            assert isinstance(result, SPResult)
            assert result.mo_path == f"path/to/mo_{conf_name}"
            assert result.energy == -100.0

    def test_execute_all_failed(
        self, mock_job_config, create_conformers, parallel_setup
    ):
        """Test execution when all jobs fail"""
        # Create two mock conformers
        conformers = create_conformers(2)
        client, cluster = parallel_setup

        # Check that RuntimeError is raised when all jobs fail
        with pytest.raises(RuntimeError, match="All jobs failed to execute"):
            execute(
                conformers=conformers,
                task=task_fail,  # type: ignore[arg-type]
                job_config=mock_job_config,
                prog=QmProg.ORCA,
                from_part="test",
                client=client,
                ignore_failed=True,
                balance=True,
                copy_mo=True,
            )

    def test_execute_partial_failure(
        self, mock_job_config, create_conformers, parallel_setup
    ):
        """Test execution with mixed success/failure"""
        # Create four mock conformers
        conformers = create_conformers(4)
        client, cluster = parallel_setup

        # Run execute
        results = execute(
            conformers=conformers,
            task=task_conditional,  # type: ignore[arg-type]
            job_config=mock_job_config,
            prog=QmProg.ORCA,
            from_part="test",
            client=client,
            ignore_failed=True,
            balance=True,
            copy_mo=True,
        )

        # Verify results (even-numbered conformers succeed, odd ones fail)
        assert len(results) == 2
        assert all(f"CONF{i}" in results for i in [2, 4])

        # Test ignore_failed
        with pytest.raises(RuntimeError, match="failed jobs"):
            results = execute(
                conformers=conformers,
                task=task_conditional,  # type: ignore[arg-type]
                job_config=mock_job_config,
                prog=QmProg.ORCA,
                from_part="test",
                client=client,
                ignore_failed=False,
                balance=True,
                copy_mo=True,
            )

    def test_execute_resource_management(
        self, mock_job_config, create_conformers, parallel_setup
    ):
        """Test resource management with multiple jobs"""
        # Create eight mock conformers
        n_jobs = 8
        conformers = create_conformers(n_jobs)
        client, cluster = parallel_setup

        # Run execute with limited cores
        start_time = time.time()
        results = execute(
            conformers=conformers,
            task=task_success,  # type: ignore[arg-type]
            job_config=mock_job_config,
            prog=QmProg.ORCA,
            from_part="test",
            client=client,
            ignore_failed=False,
            balance=False,
            copy_mo=False,
        )
        total_time = time.time() - start_time

        # Verify results
        assert len(results) == n_jobs

        # Get settings from client
        scheduler_info = client.scheduler_info()
        nnodes = int(scheduler_info["n_workers"])
        total_threads = scheduler_info["total_threads"]
        threads_per_worker = total_threads // nnodes
        total_cores = sum(
            [
                worker["resources"]["CPU"]
                for _, worker in scheduler_info["workers"].items()
            ]
        )
        ncores = total_cores // nnodes
        omp = total_cores // threads_per_worker

        # With ncores total and omp cores per job, we should be able to run ncores // omp jobs in parallel
        ncycles = ceil(n_jobs / (ncores // omp))
        min_expected_time = 0.1 * ncycles
        # Allow for some timing variability in CI environments
        max_expected_time = min_expected_time + 0.2
        assert (
            min_expected_time < total_time < max_expected_time
        ), f"Expected ~{min_expected_time:.2f}s execution time, got {total_time}s"



from typing import Self
from pydantic import model_validator, Field

from .base import BasePartConfig
from ...params import QmProg, TmSolvMod, OrcaSolvMod, GfnVersion
from ...assets import FUNCTIONALS
from ...logging import setup_logger

logger = setup_logger(__name__)


class RefinementConfig(BasePartConfig):
    """Config class for Refinement"""

    prog: QmProg = QmProg.TM
    """Program that should be used for calculations."""

    func: str = "wb97m-v"
    """Functional that should be used for calculations."""

    basis: str = "def2-TZVP"
    """Basis set that should be used for calculations."""

    sm: TmSolvMod | OrcaSolvMod = TmSolvMod.COSMORS
    """Solvation model that should be used for calculations."""

    gfnv: GfnVersion = GfnVersion.GFN2
    """GFN version to be used for xtb calculations."""

    threshold: float = Field(gt=0, le=1.0, default=0.95)
    """Boltzmann population threshold."""

    gsolv_included: bool = False
    """Whether to explicitly calculate Gsolv contributions (False) or not (True)."""

    template: bool = False
    """Whether to use template files."""

    @model_validator(mode="after")
    def func_must_be_known_in_prog(self) -> Self:
        """
        Validate that the functional is known for the chosen program.

        :return: The validated instance.
        """
        prog: str = self.prog
        try:
            assert FUNCTIONALS[self.func][prog] is not None
            assert FUNCTIONALS[self.func]["disp"] is not None
            assert FUNCTIONALS[self.func]["type"] is not None
        except (KeyError, AssertionError):
            raise ValueError(
                f"Functional {self.func} not (fully) defined for prog {prog}."
            )
        return self

    @model_validator(mode="after")
    def gsolv_included_must_be_false_for_cosmors(self) -> Self:
        """
        In case the user chose cosmors or cosmors-fine as solvation model, it might be necessary to
        switch the gsolv_included flag to False.

        :return: The validated instance.
        """
        sm: TmSolvMod | OrcaSolvMod = self.sm
        if sm in [TmSolvMod.COSMORS, TmSolvMod.COSMORS_FINE] and self.gsolv_included:
            logger.warning(
                f"Found {sm.value} as solvation model but gsolv_included is set to True. Setting to False automaticaly."
            )
            self.gsolv_included = False

        return self



import json
from pathlib import Path

from ..params import USER_ASSETS_PATH

func_path = Path(__file__).parent / "dfa.json"
user_func_path = Path(USER_ASSETS_PATH) / "dfa.json"

solv_path = Path(__file__).parent / "solvents.json"
user_solv_path = Path(USER_ASSETS_PATH) / "solvents.json"

FUNCTIONALS: dict[str, dict[str, str]] = json.loads(func_path.read_text()) | (
    json.loads(user_func_path.read_text()) if user_func_path.is_file() else {}
)
SOLVENTS: dict[str, dict[str, str]] = json.loads(solv_path.read_text()) | (
    json.loads(user_solv_path.read_text()) if user_solv_path.is_file() else {}
)



from pydantic import ValidationInfo, field_validator
from typing import Literal


from ..config.paths import PathsConfig
from .generic import GenericConfig
from ..params import (
    OrcaSolvMod,
    TmSolvMod,
    XtbSolvMod,
    GfnVersion,
    GridLevel,
)


class XTBJobConfig(GenericConfig):
    """Configuration for XTB jobs."""

    paths: PathsConfig
    """Paths to external programs."""

    gfnv: GfnVersion
    """Version of GFN to use (GFN1-xTB, GFN2-xTB, GFN-FF)."""

    gas_phase: bool
    """Whether to ignore solvation related settings."""

    solvent: str | None = None
    """Solvent name to apply. Naming in solvation model will be automatically mapped."""

    sm_rrho: XtbSolvMod | None = None
    """Solvation model to use in xtb for the mRRHO calculation."""

    @field_validator("solvent", "sm_rrho")
    @classmethod
    def need_solvent_when_no_gp(cls, v: str | None, info: ValidationInfo):
        """
        Validate that solvent and solvation model are provided if not gas-phase.

        :param v: The value to validate.
        :param info: Validation info.
        :return: The validated value.
        """
        if not info.data["gas_phase"] and v is None:
            raise ValueError(
                "Solvent and solvation model need to be provided if not doing a gas-phase calculation."
            )

        return v


class SPJobConfig(GenericConfig):
    """Configuration for single-point jobs."""

    paths: PathsConfig
    """Paths to external programs."""

    # DFT related
    copy_mo: bool
    """Whether to copy MO files for better guesses."""

    func: str
    """Functional name to use. Naming in program will be automatically mapped."""

    basis: str
    """Basis set to use."""

    grid: GridLevel
    """Grid level to be applied to this calculation."""

    template: bool
    """Whether to use template file for this calculation."""

    gas_phase: bool
    """Whether to ignore solvation related settings."""

    sm: OrcaSolvMod | TmSolvMod | None = None
    """Solvent model to use."""

    solvent: str | None = None
    """Solvent name to apply. Naming in solvation model will be automatically mapped."""

    # # Only used for COSMORS calculations
    # multitemp: bool | None = None
    # """Equivalent to general settings variant. Whether to calculate temperature range (only used in COSMORS calculations)."""

    temperature: float | None = None
    """Equivalent to general settings variant. If not multitemp then use this temperature (only used in COSMORS calculations)."""

    # trange: tuple[float, float, float] | None = None
    # """Equivalent to general settings variant. (min, max, step) for the temperature range (only used in COSMORS calculations)."""

    @field_validator("solvent", "sm")
    @classmethod
    def need_solvent_when_no_gp(cls, v: str | None, info: ValidationInfo):
        """
        Validate that solvent and solvation model are provided if not gas-phase.

        :param v: The value to validate.
        :param info: Validation info.
        :return: The validated value.
        """
        if not info.data["gas_phase"] and v is None:
            raise ValueError(
                "Solvent and solvation model need to be provided if not doing a gas-phase calculation."
            )
        return v


class OptJobConfig(SPJobConfig):
    """Configuration for optimization jobs."""

    # Optimization
    macrocycles: bool
    """Whether to use macrocycle optimization."""

    optlevel: Literal[
        "crude", "sloppy", "loose", "lax", "normal", "tight", "vtight", "extreme"
    ]
    """Optlevel for the ANCOPT optimizer. Will get mapped to roughly equivalent levels in native optimizers."""

    optcycles: int
    """Steps per macrocycles if macrocycles are used."""

    maxcyc: int | None = None
    """Maximum number of optimization steps if doing a full opt."""


class XTBOptJobConfig(OptJobConfig):
    """Configuration for XTB optimization jobs."""

    hlow: float
    """hlow setting to be used in xtb (refer to xtb documentation)."""

    constraints: str | None = None
    """Constraints in xtb format as string."""


class RRHOJobConfig(XTBJobConfig):
    """Configuration for RRHO jobs."""

    # multitemp: bool
    # """Equivalent to general settings variant. Whether to calculate temperature range (only used in COSMORS calculations)."""

    sthr: float

    imagthr: float

    # bhess: bool
    # """Whether to utilize a single-point Hessian calculation for frequencies. Otherwise geometry will be optimized (equivalent to `--bhess` vs `--ohess` for xtb)."""

    temperature: float

    # trange: tuple[float, float, float]

    # rmsdbias: bool


class NMRJobConfig(SPJobConfig):
    """Configuration for NMR jobs."""

    # NMR
    couplings: bool

    shieldings: bool

    active_nuclei: str

    fc_only: bool

    ss_cutoff: float


class RotJobConfig(SPJobConfig):
    """Configuration for optical rotation jobs."""

    freq: list[float]


class UVVisJobConfig(SPJobConfig):
    """Configuration for UVVis jobs."""

    # UVVis
    nroots: int



from typing import override
import re
from pydantic import (
    BaseModel,
    Field,
    PrivateAttr,
    field_validator,
    ConfigDict,
    ValidationInfo,
    model_validator,
)
from pathlib import Path

from ..params import PLENGTH
from ..utilities import h2


class PathsConfig(BaseModel):
    """
    Configuration for paths to external programs.
    """

    model_config = ConfigDict(
        str_strip_whitespace=True,
        use_attribute_docstrings=True,
        validate_assignment=True,
    )

    orca: str = Field("")
    """Absolute path to the orca binary."""

    _orcaversion: str = PrivateAttr("")
    """ORCA version string. Should be extracted from somewhere."""

    tm: str = Field("")
    """Absolute path to the turbomole binary directory."""

    xtb: str = Field("")
    """Absolute path to the xtb binary."""

    cosmotherm: str = Field("")
    """Absolute path to the cosmotherm binary."""

    cosmorssetup: str = Field("")
    """Name of the cosmors parameterization file."""

    @override
    def __str__(self) -> str:
        lines: list[str] = []
        lines.append(h2("PATHS of External Programs"))
        kv = [(k, v) for k, v in self]
        kv.sort(key=lambda x: type(x[1]).__name__)
        for name, value in kv:
            lines.append(f"{name:>{PLENGTH // 2 - 2}} : {value}")

        return str("\n".join(lines))

    @property
    def orcaversion(self) -> str:
        """
        Get the ORCA version string.

        :return: The ORCA version.
        """
        return self._orcaversion

    @field_validator("orca")
    def validate_orca(cls, value: str):
        """
        Validate the ORCA executable path.

        :param value: The path to validate.
        :return: The validated path.
        """
        if not Path(value).is_file():
            raise ValueError(f"orca executable not found at {value}.")
        return value

    @field_validator("tm")
    def validate_turbomole(cls, value: str):
        """
        Validate the Turbomole binary directory.

        :param value: The path to validate.
        :return: The validated path.
        """
        tmp = Path(value)
        if not all(
            (
                (tmp / "ridft").is_file(),
                (tmp / "mpshift").is_file(),
                (tmp / "escf").is_file(),
            )
        ):
            raise ValueError(
                f"Turbomole path does not contain either ridft, mpshift or escf at {value}."
            )
        return value

    @field_validator("xtb")
    def validate_xtb(cls, value: str):
        """
        Validate the XTB executable path.

        :param value: The path to validate.
        :return: The validated path.
        """
        if not Path(value).is_file():
            raise ValueError(f"xtb executable not found at {value}.")
        return value

    @field_validator("cosmotherm")
    def validate_cosmotherm(cls, value: str):
        """
        Validate the COSMOtherm executable path.

        :param value: The path to validate.
        :return: The validated path.
        """
        if not Path(value).is_file():
            raise ValueError(f"cosmotherm executable not found at {value}.")
        return value

    @field_validator("cosmorssetup")
    def validate_cosmors(cls, value: str, info: ValidationInfo):
        """
        Validate cosmotherm and cosmorssetup.

        :param value: The cosmorssetup value.
        :param info: Validation info.
        :return: The validated value.
        """
        if "cosmotherm" in info.data and info.data["cosmotherm"]:
            cosmotherm_path = Path(info.data["cosmotherm"])
            # Re-validate cosmotherm in case it was modified
            if not cosmotherm_path.is_file():
                raise ValueError(
                    f"cosmotherm executable not found at {cosmotherm_path}."
                )

            # Now validate cosmorssetup
            if value:
                setupfile = (
                    cosmotherm_path.parent / ".." / "CTDATA-FILES" / value
                ).resolve()
                if not setupfile.is_file():
                    raise ValueError(f"{setupfile} not found.")
        return value

    @model_validator(mode="after")
    def get_orca_version(self):
        """
        Extract and set the ORCA version.

        :return: The validated instance.
        """
        # if orca was found try to determine orca version from the path (kinda hacky)
        if self.orca:
            match = re.search(r"(\d+\.\d+\.\d+)", str(self.orca))
            if match:
                self._orcaversion = match.group(1)
            else:
                # Try to extract version from binary content
                with open(self.orca, "rb") as f:
                    binary_content = f.read()

                version_pattern = rb"Program Version (\d+\.\d+\.\d+)"
                match_bytes = re.search(version_pattern, binary_content)

                if not match_bytes:
                    raise ValueError("Could not determine ORCA version.")
                else:
                    version_bytes = match_bytes.group(1)
                    version_string = version_bytes.decode("utf-8")
                    self._orcaversion = version_string

        return self



import pytest
import os
from pathlib import Path
import shutil

from censo.logging import set_filehandler, set_loglevel


@pytest.fixture(autouse=True)
def tmp_wd(tmp_path: Path, monkeypatch: pytest.MonkeyPatch, request):
    orig = os.getcwd()
    if request.config.getoption("--keep-log"):
        set_filehandler(Path(orig) / "censo.log")
    set_loglevel("DEBUG")
    monkeypatch.chdir(tmp_path)
    yield
    os.chdir(orig)
    # Clean up temporary directory
    for item in tmp_path.iterdir():
        if item.is_dir():
            shutil.rmtree(item)
        else:
            item.unlink()


def pytest_addoption(parser):
    parser.addoption(
        "--run-optional", action="store_true", default=False, help="run optional tests"
    )
    parser.addoption(
        "--keep-log",
        action="store_true",
        default=False,
        help="keep the log file during test execution",
    )


def pytest_collection_modifyitems(config, items):
    if config.getoption("--run-optional"):
        # --run-optional given in cli: do not skip optional tests
        return
    skip_optional = pytest.mark.skip(reason="need --run-optional option to run")
    for item in items:
        if "optional" in item.keywords:
            item.add_marker(skip_optional)



from dataclasses import fields
import json
from collections import defaultdict
from collections.abc import Mapping
import pytest
from pathlib import Path

from censo.ensemble import EnsembleData
from censo.processing.results import (
    SPResult,
    OptResult,
    GsolvResult,
    RRHOResult,
    NMRResult,
    UVVisResult,
)


def tree():
    return defaultdict(tree)


def to_plain_dict(d):
    if isinstance(d, Mapping):
        return {k: to_plain_dict(v) for k, v in d.items()}
    return d


# ============= Fixtures and Mock Classes =============


@pytest.fixture
def mock_ensemble(fixtures_path: Path):
    """Create a glycerol ensemble"""
    ensemble = EnsembleData()
    ensemble.read_input((fixtures_path / "crest_conformers.xyz").as_posix())

    return ensemble


@pytest.fixture
def mock_execute_results(fixtures_path: Path):
    """Load real results for the glycerol ensemble from fixtures"""
    # Load data
    prescreening = json.loads((fixtures_path / "0_PRESCREENING.json").read_text())[
        "results"
    ]
    screening = json.loads((fixtures_path / "1_SCREENING.json").read_text())["results"]
    optimization = json.loads((fixtures_path / "2_OPTIMIZATION.json").read_text())[
        "results"
    ]
    refinement = json.loads((fixtures_path / "3_REFINEMENT.json").read_text())[
        "results"
    ]
    nmr = json.loads((fixtures_path / "4_NMR.json").read_text())["results"]
    uvvis = json.loads((fixtures_path / "6_UVVIS.json").read_text())["results"]

    # Collect
    results = tree()
    for conf in prescreening:
        res_sp = SPResult(**prescreening[conf]["sp"])
        results["prescreening"]["sp"][conf] = res_sp
        res_gsolv = GsolvResult(**prescreening[conf]["xtb_gsolv"])
        results["prescreening"]["xtb_gsolv"][conf] = res_gsolv

    for conf in screening:
        res_gsolv = GsolvResult(**screening[conf]["gsolv"])
        results["screening"]["gsolv"][conf] = res_gsolv
        res_sp = SPResult("", res_gsolv.energy_gas)
        results["screening"]["sp"][conf] = res_sp
        res_rrho = RRHOResult(**screening[conf]["xtb_rrho"])
        results["screening"]["xtb_rrho"][conf] = res_rrho

    for conf in optimization:
        res_opt = OptResult(**optimization[conf]["opt"])
        results["optimization"]["opt"][conf] = res_opt
        res_rrho = RRHOResult(**optimization[conf]["xtb_rrho"])
        results["optimization"]["xtb_rrho"][conf] = res_rrho

    for conf in refinement:
        res_sp = SPResult(**refinement[conf]["sp"])
        results["refinement"]["sp"][conf] = res_sp
        res_rrho = RRHOResult(**refinement[conf]["xtb_rrho"])
        results["refinement"]["xtb_rrho"][conf] = res_rrho

    for conf in nmr:
        res_nmr = NMRResult(
            **{k: v for k, v in nmr[conf]["nmr"].items() if k in fields(NMRResult)}
        )
        results["nmr"][conf] = res_nmr

    for conf in uvvis:
        res_uvvis = UVVisResult(
            **{
                k: v
                for k, v in uvvis[conf]["uvvis"].items()
                if k in fields(UVVisResult)
            }
        )
        results["uvvis"][conf] = res_uvvis

    return to_plain_dict(results)



"""
Utility functions which are used in the CENSO modules. From creating folders to
printout routines.
"""

import functools
import os
import time
from collections.abc import Callable
from pydantic import BaseModel
from typing import Any
from datetime import timedelta

from .params import BOHR2ANG, PLENGTH
from .logging import setup_logger
import json
from pydantic import ValidationError

logger = setup_logger(__name__)


def print_validation_errors(e: ValidationError) -> None:
    """
    Print Pydantic validation errors in a human-readable format.

    :param e: The ValidationError instance containing the errors.
    :return: None
    """
    printf(f"Found {e.error_count()} validation error(s):\n")
    printf(f"Found {e.error_count()} validation error(s):\n")
    for error in e.errors():
        field = " -> ".join(map(str, error["loc"]))
        message = error["msg"]
        user_input = error["input"]
        # Handle model-level validator errors differently
        if not error["loc"] or (
            len(error["loc"]) == 1 and error["loc"][0] == "__root__"
        ):
            printf("  - Model-level error:")
            printf(f"    Message: {message}")
        else:
            try:
                user_input_str = json.dumps(user_input)
            except TypeError:
                user_input_str = str(user_input)
            printf(f"  - Field: '{field}'")
            printf(f"    Message: {message}")
            printf(f"    Your input: {user_input_str}")
        printf("-" * 20)


class Factory[T]:
    """
    Abstract object factory class.
    """

    _registry: dict[str, Callable[..., T]] = {}

    @classmethod
    def register_builder(cls, name: str, builder: Callable[..., T]) -> None:
        """
        Register a builder.

        :param name: Name of the builder.
        :param builder: The builder callable.
        :return: None
        """
        cls._registry[name] = builder
        cls._registry[name] = builder

    @classmethod
    def create(cls, name: str, *args, **kwargs) -> T:
        """
        Generic factory method.

        :param name: Name of the builder to create.
        :return: The created instance.
        """
        if name not in cls._registry:
            raise TypeError(f"No type was found for '{name}' in {list(cls._registry)}.")
        builder = cls._registry[name]
        return builder(*args, **kwargs)


class DataDump(BaseModel):
    """
    Model for dumping data with part name and settings.
    """

    part_name: str
    data: dict[
        str,
        dict[
            str,
            Any,
        ],
    ] = {}
    settings: dict[str, Any] = {}


def printf(*args, **kwargs):
    """
    patch print to always flush
    """
    print(*args, flush=True, **kwargs)


def frange(start: float, end: float, step: float = 1) -> list[float]:
    """
    Create a range of floats, adding 'step' to 'start' while it's less or equal than 'end'.

    :param start: The start of the range.
    :param end: The end of the range.
    :param step: The step size. Defaults to 1.
    :return: The list of floats.
    """
    result = []
    current = start
    while current <= end:
        result.append(current)
        current += step
    return result


def t2x(
    path: str, writexyz: bool = False, outfile: str = "original.xyz"
) -> tuple[list[str], int, str]:
    """
    Convert TURBOMOLE coord file to xyz data and/or write *.xyz output.

    :param path: Absolute path either to dir or file directly.
    :param writexyz: If True, directly write to outfile. Defaults to False.
    :param outfile: Filename of xyz file which is written into the same directory as path.
    :return: Tuple of (coordxyz list of strings, number of atoms, output path).
    """
    # read lines from coord file
    with open(path, newline=None) as f:
        coord = f.readlines()

    # read coordinates with atom labels directly into a string
    # and append the string to a list to be written/returned later
    xyzatom = []
    for line in coord:
        if "$end" in line:  # stop at $end ...
            break
        xyzatom.append(
            functools.reduce(
                lambda x, y: x + " " + y,
                [
                    f"{float(line.split()[0]) * BOHR2ANG:.10f}",
                    f"{float(line.split()[1]) * BOHR2ANG:.10f}",
                    f"{float(line.split()[2]) * BOHR2ANG:.10f}",
                    f"{str(line.split()[3].lower()).capitalize()}",
                ],
            )
        )

    # get path from args without the filename of the ensemble (last element of path)
    if os.path.isfile(path):
        outpath = functools.reduce(
            lambda x, y: os.path.join(x, y), list(os.path.split(path))[::-1][1:][::-1]
        )
    # or just use the given path if it is not a file path
    else:
        outpath = path

    # write converted coordinates to xyz outfile if wanted
    if writexyz:
        with open(os.path.join(outpath, outfile), "w") as out:
            out.write(str(len(xyzatom)) + "\n")
            for line in xyzatom:
                out.write(line)
    return xyzatom, len(xyzatom), os.path.join(outpath, outfile)


def check_for_float(line: str) -> float | None:
    """
    Go through line and check for float, return first float.

    :param line: The line to check.
    :return: The first float found, or None.
    """
    elements = line.strip().split()
    value = None
    for element in elements:
        try:
            value = float(element)
        except ValueError:
            value = None
        if value is not None:
            break
    return value


def timeit(f: Callable[..., None]) -> Callable[..., float]:
    """
    Time function execution.

    .. deprecated::
        This decorator is deprecated. Timing should be handled at the CLI level.
        TODO: Remove in a follow-up change.

    :param f: The function to time (should have no return value).
    :return: A wrapper that returns the execution time.
    """

    @functools.wraps(f)
    def wrapper(*args, **kwargs) -> float:
        start = time.perf_counter()
        f(*args, **kwargs)
        end = time.perf_counter()
        return end - start

    return wrapper


def get_time(time: float) -> tuple[int, int, int]:
    """
    Calculate seconds, minutes, hours from time in seconds.

    :param time: The time in seconds.
    :return: Tuple of seconds, minutes, hours.
    """
    time_taken = timedelta(seconds=int(time))
    hours, r = divmod(time_taken.seconds, 3600)
    minutes, seconds = divmod(r, 60)
    if time_taken.days:
        hours += time_taken.days * 24
    return seconds, minutes, hours


def h1(text: str) -> str:
    """
    Create a formatted header of type 1.

    :param text: The text to be formatted.
    :return: The formatted header.
    """
    return "\n" + f" {text} ".center(PLENGTH, "-") + "\n"


def h2(text: str) -> str:
    """
    Create a formatted header of type 2.

    :param text: The text to be formatted.
    :return: The formatted header.
    """
    return f"""
{"-" * PLENGTH}
{text.center(PLENGTH, " ")}
{"-" * PLENGTH}
    """



"""
Contains configuration related classes and functions.

Important note for configuration in general:
Many of the models created with pydantic have field_validators that check other already
validated values to validate certain settings. If you set a setting in a way that (temporarily)
an invalid configuration would result, this will raise an error. If you want to avoid this,
e.g. for setting the functional to something not available for the program you want to use
you need to disable assignment validation manually: <model>.model_config.validate_assignment = False.
"""

from .generic import GenericConfig
from .parts_config import PartsConfig
from .parts import (
    PrescreeningConfig,
    ScreeningConfig,
    OptimizationConfig,
    RefinementConfig,
    NMRConfig,
    UVVisConfig,
    GeneralConfig,
    BasePartConfig,
)

__all__ = [
    "PartsConfig",
    "PrescreeningConfig",
    "ScreeningConfig",
    "OptimizationConfig",
    "RefinementConfig",
    "NMRConfig",
    "UVVisConfig",
    "GeneralConfig",
    "BasePartConfig",
    "GenericConfig",
]



from .general import GeneralConfig
from .prescreening import PrescreeningConfig
from .screening import ScreeningConfig
from .optimization import OptimizationConfig
from .refinement import RefinementConfig
from .nmr import NMRConfig
from .uvvis import UVVisConfig
from .base import BasePartConfig
from .rot import RotConfig

__all__ = [
    "GeneralConfig",
    "PrescreeningConfig",
    "ScreeningConfig",
    "OptimizationConfig",
    "RefinementConfig",
    "NMRConfig",
    "UVVisConfig",
    "BasePartConfig",
    "RotConfig",
]



import logging
import sys
from pathlib import Path

__loglevel = logging.INFO

# _loglevel = logging.DEBUG


def setup_logger(name: str) -> logging.Logger:
    """
    Initializes and configures a logger with the specified name.

    :param name: The name of the logger.
    :return: The configured logger instance.
    """
    # Create a logger instance with the specified name
    logger = logging.getLogger(name)
    logger.setLevel(__loglevel)

    # Only add handlers if the logger doesn't have any
    if not logger.handlers:
        # Create a StreamHandler to log messages to stdout (only WARNING or above)
        stream_handler = logging.StreamHandler(sys.stdout)
        stream_handler.setLevel(logging.WARNING)

        # Define the log message format
        stream_formatter = logging.Formatter("{levelname:<10s}- {message}", style="{")
        stream_handler.setFormatter(stream_formatter)

        # Add the StreamHandler to the logger
        logger.addHandler(stream_handler)

    return logger


def set_filehandler(path: str | Path):
    """
    Set filehandler for all censo loggers, avoiding duplicates.

    :param path: Path to the log file.
    :return: None
    """
    filehandler_path = str(path)
    formatter = logging.Formatter(
        "{asctime:24s}-{name:^24s}-{levelname:^10s}- {message}", style="{"
    )
    # Loop over all loggers registered in logging
    for logger_name, logger in logging.Logger.manager.loggerDict.items():
        if isinstance(logger, logging.Logger) and logger_name.startswith("censo"):
            # Check for pre-existing FileHandlers with same path
            filehandler_exists = any(
                isinstance(h, logging.FileHandler)
                and getattr(h, "baseFilename", None) == filehandler_path
                for h in logger.handlers
            )
            if not filehandler_exists:
                handler = logging.FileHandler(path)
                handler.setFormatter(formatter)
                logger.addHandler(handler)


def set_loglevel(loglevel: str) -> None:
    """
    Set the log level for all censo loggers.

    :param loglevel: The log level to set.
    :return: None
    """
    global __loglevel
    __loglevel = getattr(logging, loglevel)
    for logger_name, logger in logging.Logger.manager.loggerDict.items():
        if isinstance(logger, logging.Logger) and logger_name.startswith("censo"):
            logger.setLevel(__loglevel)



from typing import Literal, Self
from pydantic import field_validator, Field, model_validator

from .base import BasePartConfig
from ...params import QmProg, TmSolvMod, OrcaSolvMod
from ...assets import FUNCTIONALS


class NMRConfig(BasePartConfig):
    """Config class for NMR"""

    prog: QmProg = QmProg.ORCA
    """Program that should be used for calculations."""

    func: str = "pbe0-d4"
    """Functional that should be used for calculations."""

    basis: str = "def2-tzvp"
    """Basis set that should be used for calculations."""

    sm: Literal[TmSolvMod.COSMO, TmSolvMod.DCOSMORS] | OrcaSolvMod = OrcaSolvMod.SMD
    """Solvation model that should be used for calculations."""

    resonance_frequency: float = Field(gt=0, default=300.0)
    """Resonance frequency assumed for calculation of coupling constants."""

    ss_cutoff: float = Field(gt=0, default=8.0)
    """Only for ORCA: passed to SpinSpinThresh keyword."""

    fc_only: bool = True
    """Whether to only calculate the Fermi-Contact term for spin-spin couplings."""

    shieldings: bool = True
    """Whether to calculate shieldings."""

    couplings: bool = True
    """Whether to calculate spin-spin couplings."""

    active_nuclei: str = "h,c"
    """Nuclei active for NMR calculations."""

    template: bool = False
    """Whether to use template files."""

    @model_validator(mode="after")
    def func_must_be_known_in_prog(self) -> Self:
        """
        Validate that the functional is known for the chosen program.

        :return: The validated instance.
        """
        prog: str = self.prog
        try:
            assert FUNCTIONALS[self.func][prog] is not None
            assert FUNCTIONALS[self.func]["disp"] is not None
            assert FUNCTIONALS[self.func]["type"] is not None
        except (KeyError, AssertionError):
            raise ValueError(
                f"Functional {self.func} not (fully) defined for prog {prog}."
            )
        return self

    @field_validator("active_nuclei")
    @classmethod
    def active_nuclei_must_be_available(cls, v: str):
        """
        Validate that active nuclei are available.

        :param v: The nuclei string.
        :return: The validated nuclei string.
        """
        nuclei = [s.lower() for s in v.split(",")]
        if not all(n in ["h", "c", "f", "si", "p"] for n in nuclei):
            raise ValueError(
                f"Selection of active nuclei ({v}) invalid. Allowed nuclei: h, c, f, si, p."
            )
        return v



from inspect import ismethod
import traceback
import os
from typing import Literal, Callable
from concurrent.futures import CancelledError
from dask.distributed import (
    Variable,
    Client,
    LocalCluster,
    as_completed as dask_as_completed,
)


from .molecules import MoleculeData
from .logging import setup_logger
from .params import OMPMAX_DEFAULT, QmProg, OMPMIN_DEFAULT
from .config.job_config import XTBJobConfig, SPJobConfig
from .processing.results import Result, MetaData
from .processing.job import JobContext
from .config import GenericConfig


logger = setup_logger(__name__)


def get_cluster(maxcores: int | None = None, ompmin: int = OMPMIN_DEFAULT):
    """
    Set up Dask parallel execution environment.

    :param maxcores: Total number of cores made available for the LocalCluster.
    :type maxcores: int | None
    :param ompmin: Minimum number of cores per job.
    :type ompmin: int
    :return: Yields LocalCluster.
    :rtype: LocalCluster
    """
    nnodes = int(os.environ.get("SLURM_NNODES", "1"))
    slurm_ntasks_str = os.environ.get("SLURM_NTASKS", None)
    slurm_ntasks = None

    if slurm_ntasks_str is not None:
        slurm_ntasks = int(slurm_ntasks_str)
        if nnodes > 1:
            slurm_ntasks = slurm_ntasks // nnodes
        logger.debug(f"Found SLURM_NTASKS={slurm_ntasks} (for each node).")

    if maxcores is None:
        ncores = slurm_ntasks if slurm_ntasks is not None else os.cpu_count()
        if ncores is None:
            raise RuntimeError("Could not determine number of available cores.")
    elif slurm_ntasks is not None and maxcores <= slurm_ntasks:
        ncores = maxcores
    elif slurm_ntasks is not None and maxcores > slurm_ntasks:
        raise RuntimeError(
            "Requested more cores than slurm tasks available for this node."
        )
    else:
        ncores = maxcores

    # TODO: SSHCluster for multiple nodes
    logger.debug(f"Setting up LocalCluster with {ncores} cores.")
    cluster = LocalCluster(
        n_workers=1,
        threads_per_worker=ncores // ompmin,
        resources={"CPU": ncores},  # One worker receives all CPUs
    )
    return cluster


def set_omp(
    jobs: list[JobContext],
    balance: bool,
    omp: int,
    ncores: int,
):
    """
    Sets the number of cores to be used for every job.

    :param jobs: List of parallel jobs.
    :type jobs: list[JobContext]
    :param balance: Whether to balance.
    :type balance: bool
    :param omp: Default OMP value.
    :type omp: int
    :param ncores: Total cores.
    :type ncores: int
    :return: None
    :rtype: None
    """
    if balance:
        jobs_left, tot_jobs = len(jobs), len(jobs)

        if tot_jobs >= ncores // omp:
            # Assign minimum OMP for as many jobs as possible
            for i in range(tot_jobs - tot_jobs % omp):
                jobs[i].omp = omp

            jobs_left = tot_jobs % omp

        if jobs_left > 0:
            if jobs_left == 1:
                jobs[-1].omp = ncores
            else:
                while jobs_left > 0:
                    omps = list(range(omp, min(OMPMAX_DEFAULT, ncores) + 1))
                    jobs_in_parallel = {o: ncores // o for o in omps}

                    # Remove OMP values that would lead to idle groups
                    omps = [o for o in omps if jobs_left - jobs_in_parallel[o] >= 0]

                    divisors = [o for o in omps if ncores % o == 0 and o != ncores]
                    if len(divisors) > 0:
                        tmp_omp = min(divisors)
                    else:
                        # Not possible to evenly divide cores
                        # Pick OMP value with least idle cores that is not ncores
                        # Chances are that it is one of the smallest OMP values
                        omps = [o for o in omps if o != ncores]
                        if len(omps) > 0:
                            tmp_omp = min(omps, key=lambda x: ncores % x)
                        else:
                            tmp_omp = ncores

                    jobs[tot_jobs - jobs_left].omp = tmp_omp
                    jobs_left -= jobs_in_parallel[tmp_omp]
    else:
        for job in jobs:
            job.omp = omp


def prepare_jobs(
    conformers: list[MoleculeData],
    prog: str,
    ncores: int,
    omp: int,
    from_part: str,
    balance: bool = True,
    copy_mo: bool = True,
) -> list[JobContext]:
    """
    Prepares the jobs from the conformers data.

    :param conformers: List of conformers.
    :type conformers: list[MoleculeData]
    :param prog: Program name.
    :type prog: str
    :param ncores: Total cores per node.
    :type ncores: int
    :param omp: Default OMP value.
    :type omp: int
    :param from_part: From part.
    :type from_part: str
    :param balance: Whether to balance.
    :type balance: bool
    :param copy_mo: Whether to copy MO.
    :type copy_mo: bool
    :return: List of parallel jobs.
    :rtype: list[JobContext]
    """
    # Create JobContext instances from the conformers, sharing the prepinfo dict
    jobs = [
        JobContext(conf.geom, conf.charge, conf.unpaired, omp) for conf in conformers
    ]
    if copy_mo:
        for job in jobs:
            # Insert mo guess file path
            if prog == "orca":
                job.mo_guess = next(
                    (
                        mo_path
                        for conf in conformers
                        for mo_path in conf.mo_paths["orca"]
                        if conf.name == job.conf.name and ".gbw" in mo_path
                    ),
                    None,
                )
            elif prog == "tm":
                job.mo_guess = next(
                    (
                        mo_path
                        for conf in conformers
                        for mo_path in conf.mo_paths["tm"]
                        if conf.name == job.conf.name
                        and any(kw in mo_path for kw in ["alpha", "beta", "mos"])
                    ),
                    None,
                )
            job.from_part = from_part

    if balance:
        set_omp(jobs, balance, omp, ncores)

    return sorted(jobs, key=lambda job: job.omp)


def execute[
    T: GenericConfig, U: Result
](
    conformers: list[MoleculeData],
    task: Callable[[JobContext, T], tuple[U, MetaData]],
    job_config: XTBJobConfig | SPJobConfig,
    prog: QmProg | Literal["xtb"],
    from_part: str,
    client: Client,
    ignore_failed: bool = True,
    balance: bool = True,
    copy_mo: bool = True,
) -> dict[str, U]:
    """
    Executes the parallel tasks using Dask distributed execution.

    :param conformers: List of conformers for which jobs will be created and executed.
    :type conformers: list[MoleculeData]
    :param task: Callable to be mapped onto the list of jobs created from conformers. This should always be a processor function.
    :type task: Callable[[JobContext, T], tuple[U, MetaData]]
    :param job_config: Instructions for the execution of the task.
    :type job_config: XTBJobConfig | SPJobConfig
    :param prog: Name of the QM program.
    :type prog: QmProg | Literal["xtb"]
    :param from_part: From part.
    :type from_part: str
    :param client: dask.distributed.Client for parallel execution.
    :type client: Client
    :param ignore_failed: Whether to ignore failed jobs.
    :type ignore_failed: bool
    :param balance: Whether to balance the number of cores used per job.
    :type balance: bool
    :param copy_mo: Whether to store the paths to the MO files for reuse.
    :type copy_mo: bool
    :return: Job results.
    :rtype: dict[str, U]
    """
    # Setting the collective cancellation variable
    cancel_var = None
    if ismethod(task):
        # Only meaningful if the task is a method of a GenericProc
        # but this is difficult to annotate with the present architecture
        proc = task.__self__
        if hasattr(proc, "cancel_var"):
            proc_id = id(proc)

            # Cooperative cancellation variable
            cancel_var_name = f"censo_cancel_{proc_id}"
            cancel_var = Variable(cancel_var_name, client=client)
            cancel_var.set(False)
            setattr(proc, "cancel_var", cancel_var)

    logger.debug("Using Dask parallel environment...")

    # Get settings from client
    scheduler_info = client.scheduler_info()
    nnodes = int(scheduler_info["n_workers"])
    total_threads = scheduler_info["total_threads"]
    threads_per_worker = total_threads // nnodes
    total_cores = sum(
        [worker["resources"]["CPU"] for _, worker in scheduler_info["workers"].items()]
    )
    ncores = total_cores // nnodes
    omp = total_cores // threads_per_worker

    # Prepare jobs for parallel execution by assigning number of cores per job etc.
    jobs: list[JobContext] = prepare_jobs(
        conformers,
        prog,
        ncores,
        omp,
        from_part,
        balance=balance,
        copy_mo=copy_mo,
    )

    # Submit jobs directly
    futures = []
    for job in jobs:
        resources = {"CPU": int(job.omp)}
        fut = client.submit(
            task,
            job,
            job_config,
            resources=resources,
        )
        futures.append(fut)

    failed_confs: list[MoleculeData] = []
    results: dict[str, U] = {}
    broken = False

    logger.debug("Waiting for jobs to complete...")
    try:
        completed_iter = dask_as_completed(futures)

        # Process results as they come in
        for completed_future in completed_iter:
            conf_name = None
            try:
                result, meta = completed_future.result()
                conf_name = meta.conf_name
                conf = next(c for c in conformers if c.name == conf_name)
                if not meta.success:
                    failed_confs.append(conf)
                    logger.warning(
                        f"{meta.conf_name} job failed. Error: {meta.error}. Check output files."
                    )
                else:
                    results[meta.conf_name] = result
                    if prog in QmProg:
                        conf.mo_paths[prog].append(result.mo_path)
            except CancelledError:
                logger.debug("Future cancelled.")
                broken = True
            except Exception:
                # Fatal error: set cancel flag and best-effort cancel all remaining futures
                logger.critical(
                    "Encountered exception in task. Cancelling remaining tasks."
                )
                if cancel_var is not None:
                    cancel_var.set(True)
                try:
                    client.cancel(futures)
                except Exception:
                    logger.debug(
                        "client.cancel raised while trying to cancel futures.",
                        exc_info=True,
                    )
                tb = traceback.format_exc()
                logger.debug(tb)
                # Try to identify failure
                if conf_name is None:
                    try:
                        idx = futures.index(completed_future)
                        if 0 <= idx < len(jobs):
                            conf_name = jobs[idx].conf.name
                    except Exception:
                        pass
                logger.debug(f"Job failed for {conf_name}:\n{tb}")
                broken = True
                break
    except KeyboardInterrupt:
        logger.warning("Received keyboard interrupt. Cancelling remaining tasks...")
        if cancel_var is not None:
            cancel_var.set(True)
        try:
            client.cancel(futures)
        except Exception:
            logger.debug(
                "client.cancel raised during KeyboardInterrupt handling.", exc_info=True
            )
        raise

    if broken:
        raise RuntimeError("Exception encountered in parallel execution.")

    # Summary
    if len(failed_confs) == len(conformers):
        raise RuntimeError(
            "All jobs failed to execute. Please check your setup and output files.\n"
        )

    if len(failed_confs) > 0:
        logger.info(f"Number of failed jobs: {len(failed_confs)}.")
        logger.info("Failed conformers:")
        logger.info(", ".join(c.name for c in failed_confs))
        if not ignore_failed:
            raise RuntimeError(f"Stopping CENSO due to failed jobs in {from_part}.")
    else:
        logger.info("All jobs executed successfully.")

    return results



import pytest
from censo.config.parts.prescreening import PrescreeningConfig
from censo.params import GfnVersion, QmProg


def test_prescreening_config_default_values():
    """Test default values of PrescreeningConfig"""
    config = PrescreeningConfig()

    assert config.prog == QmProg.TM
    assert config.func == "pbe-d3"  # should be changed to d3 from d4 because of GCP bug
    assert config.basis == "def2-sv(p)"
    assert config.gfnv == GfnVersion.GFN2
    assert config.threshold == 4.0
    assert config.template is False


def test_threshold_validation():
    """Test threshold validation"""
    with pytest.raises(ValueError):
        PrescreeningConfig(threshold=0.0)
    with pytest.raises(ValueError):
        PrescreeningConfig(threshold=-1.0)


@pytest.mark.parametrize(
    "prog,func,should_pass",
    [
        (QmProg.TM, "pbe-d3", True),
        (QmProg.ORCA, "pbe-d4", True),  # Assuming this is valid
        (QmProg.TM, "invalid-func", False),
    ],
)
def test_functional_validation(prog, func, should_pass):
    """Test functional validation with different combinations"""
    if should_pass:
        config = PrescreeningConfig(prog=prog, func=func)
        assert config.prog == prog
        assert config.func == func
    else:
        with pytest.raises(ValueError):
            PrescreeningConfig(prog=prog, func=func)



"""
stores ensembledata and conformers
functionality for program setup
"""

from pathlib import Path
import re
import json
from typing import Callable, Iterator
from math import exp

from .molecules import MoleculeData, Contributions
from .logging import setup_logger
from .params import AU2J, KB
from .utilities import check_for_float, printf, t2x

logger = setup_logger(__name__)


class EnsembleData:
    """
    Class to store conformer rotamer ensembles for use in CENSO.
    """

    def __init__(self):
        """
        Setup an EnsembleData object. Conformers can be read using read_input.
        """
        # stores the conformers with all info
        self.__conformers: list[MoleculeData] = []

        # stores the conformers which were sorted out
        self.__rem: list[MoleculeData] = []

        # String containing the constraints in xtb format to be applied to geometry optimizations
        # TODO: and xtb_rrho?
        self.constraints: str | None = None

    def __iter__(self) -> Iterator[MoleculeData]:
        return iter(self.__conformers)

    @property
    def conformers(self):
        """
        Returns the conformers list. Includes a check whether there are any conformers left.

        :return: List of conformers.
        """
        # TODO - no checks for now
        return self.__conformers

    @conformers.setter
    def conformers(self, confs: list[MoleculeData]):
        """
        Set the conformers list.

        :param confs: List of MoleculeData instances.
        :type confs: list[MoleculeData]
        :raises ValueError: If not all objects are MoleculeData instances.
        """
        if not all(isinstance(conf, MoleculeData) for conf in confs):
            raise ValueError("All objects need to be MoleculeData instances.")
        self.__conformers = confs

    @property
    def rem(self) -> list[MoleculeData]:
        """
        Returns the list of removed conformers.

        :return: List of removed conformers.
        :rtype: list[MoleculeData]
        """
        return self.__rem

    def read_output(self, outpath: str | Path) -> None:
        """
        Read json output file of a previous execution. Will try to load data into current conformer ensemble, matching
        based on names. If a conformer name does not exist in the current ensemble it will be ignored. If a conformer
        does not exist in the output data RuntimeError will be raised.

        :param outpath: Path to the output file.
        :type outpath: str | Path
        :return: None
        :rtype: None
        """

        data = json.loads(Path(outpath).read_text())

        # Check if all conformers from the current ensemble are also found in the output data
        if not all(conf.name in data for conf in self.conformers):
            raise RuntimeError(
                "Not all conformers from the current ensemble are found in the output data."
            )

        for conf in self.conformers:
            contributions = Contributions(
                data[conf.name]["energy"],
                data[conf.name]["gsolv"],
                data[conf.name]["grrho"],
            )
            conf.update(contributions)

        logger.info(f"Reloaded results from {outpath}.")

    def read_input(
        self,
        input_path: str,
        charge: int = 0,
        unpaired: int = 0,
        nconf: int | None = None,
        append: bool = False,
    ) -> None:
        """
        Read ensemble input file. Should be a file in xyz-file format with all the conformers in consecutive order.

        :param input_path: Path to the ensemble input file.
        :type input_path: str
        :param charge: Sets the charge of all molecules to this value. Defaults to 0.
        :type charge: int
        :param unpaired: Sets the unpaired electrons of all molecules to this value. Defaults to 0.
        :type unpaired: int
        :param nconf: Number of conformers to consider. Defaults to None, so all conformers are read.
        :type nconf: int | None
        :param append: If True, the conformers will be appended to the existing ensemble. Defaults to False.
        :type append: bool
        :return: None
        :rtype: None
        """
        # If $coord in file => tm format, needs to be converted to xyz
        with open(input_path) as inp:
            lines = inp.readlines()
            if any("$coord" in line for line in lines):
                _, _, input_path = t2x(
                    input_path, writexyz=True, outfile="converted.xyz"
                )

        confs = self.__setup_conformers(input_path)
        if len(confs) == 0:
            logger.warning("Input file is empty!")

        if nconf is None:
            nconf = len(confs)

        if append:
            self.conformers.extend(confs[:nconf])
        else:
            self.conformers = confs[:nconf]

        try:
            self.conformers.sort(key=lambda x: x.xtb_energy)
        except TypeError:
            # Only sort if all conformers have a defined precalculated energy
            pass

        for conf in self.conformers:
            conf.charge = charge
            conf.unpaired = unpaired

        # Print information about read ensemble
        printf(
            f"Read {len(self.conformers)} conformers.\n\n",
        )

    def get_populations(self, temperature: float) -> dict[str, float]:
        """
        Calculate populations for boltzmann distribution of ensemble at given
        temperature given values for free enthalpy.

        :param temperature: Temperature for Boltzmann distribution.
        :type temperature: float
        :return: Dictionary of conformer names to populations.
        :rtype: dict[str, float]
        """
        # find lowest gtot value
        minfree: float = min(conf.gtot for conf in self.conformers)

        # calculate boltzmann factors
        bmfactors = {
            conf.name: conf.degen
            * exp(-(conf.gtot - minfree) * AU2J / (KB * temperature))
            for conf in self.conformers
        }

        # calculate partition function from boltzmann factors
        bsum: float = sum(bmfactors.values())

        # Set Boltzmann populations
        populations: dict[str, float] = {}
        for conf in self.conformers:
            populations[conf.name] = bmfactors[conf.name] / bsum

        return populations

    def update_contributions(self, contributions_dict: dict[str, Contributions]):
        """
        Update contributions for all conformers in the ensemble. Convenience wrapper to call update on all conformers with correct mapping.

        :param contributions_dict: Dictionary mapping conformer names to Contributions.
        :type contributions_dict: dict[str, Contributions]
        :raises AssertionError: If contributions dict is incomplete.
        """
        assert all(
            conf.name in contributions_dict for conf in self.conformers
        ), "Attempted to pass incomplete contributions dict to update_contributions."
        for conf in self.conformers:
            conf.update(contributions_dict[conf.name])

    def __setup_conformers(self, input_path: str) -> list[MoleculeData]:
        """
        Open ensemble input, split into conformers, create MoleculeData objects out of coord input, read out energy from xyz file if possible.
        In principle this can also read xyz-files with molecules of different sizes.

        :param input_path: Path to the ensemble input file.
        :type input_path: str
        :return: A list of MoleculeData objects.
        :rtype: list[MoleculeData]
        """
        # open ensemble input
        with open(input_path) as file:
            lines = file.readlines()

        # Get rid of unnecessary empty lines
        # Basically this filters out all only-whitespace lines except the comment lines after the number of atoms is declared
        lines = list(
            filter(
                lambda line: not (
                    bool(re.match(r"^\s*$", line))  # matches only whitespace chars
                    and len(lines[lines.index(line) - 1].split()) != 1
                ),
                lines,
            )
        )

        # assuming consecutive xyz-file format
        # (every conf geometry is separated by a line with split length of 4 followed by a line of split length 1)
        #
        # 14                               <-- split_index refers to this line (this is line 0 for the first conf)
        # CONF12 -22.521386
        # H x.xxxxxxxx x.xxxxxxx x.xxxxxx
        # ...
        split_indices = [
            i
            for i in range(len(lines))
            if i == 0 or (len(lines[i].split()) == 1 and len(lines[i - 1].split()) == 4)
        ]

        conformers = []
        for i, split_index in enumerate(split_indices):
            # Check whether the names are stored in the ensemble file,
            # use those if possible because of crest rotamer files
            if "CONF" in lines[split_index + 1]:
                confname = next(
                    s for s in lines[split_index + 1].split() if "CONF" in s
                )
            else:
                # Start counting from 1
                confname = f"CONF{i + 1}"

            # Determine end of geometry definition for this conf
            # which is either the next conf definition or EOF
            conf_end_index = (
                split_indices[i + 1] if i + 1 < len(split_indices) else len(lines)
            )

            # Create a new conformer object and append it to the ensemble
            conformers.append(
                MoleculeData(
                    confname,
                    lines[split_index + 2 : conf_end_index],
                )
            )

            # get precalculated energies if possible
            # precalculated energy set to 0.0 if it cannot be found
            conformers[i].xtb_energy = check_for_float(lines[split_index + 1]) or 0.0

        return conformers

    def remove_conformers(self, cond: Callable[[MoleculeData], bool]) -> None:
        """
        Remove conformers from further consideration if 'cond' evaluates to True.
        The removed conformers will be stored in self.rem.

        :param cond: Condition to check for the conf objects.
        :type cond: Callable[[MoleculeData], bool]
        :return: None
        :rtype: None
        """
        filtered = list(filter(cond, self.conformers))
        for conf in filtered:
            # pop item from conformers and insert this item at index 0 in rem
            self.rem.insert(0, self.conformers.pop(self.conformers.index(conf)))

            # Log removed conformers
            logger.debug(f"Removed {conf.name}.")

    def dump_xyz(self, file: Path):
        """
        Dump the current ensemble in xyz-format.

        :param file: Path to write the xyz file.
        :type file: Path
        :return: None
        :rtype: None
        """
        text = "".join(sum([conf.geom.toxyz() for conf in self], []))
        file.write_text(text)

    def dump_rem_xyz(self, file: Path):
        """
        Dump the conformers removed via 'remove_conformers' in xyz-format.

        :param file: Path to write the xyz file.
        :type file: Path
        :return: None
        :rtype: None
        """
        text = "".join(sum([conf.geom.toxyz() for conf in self.rem], []))
        file.write_text(text)

    def dump_json(self, file: Path):
        """
        Dump the ensemble with most recent rankings and values in json-format.

        :param file: Path to write the json file.
        :type file: Path
        :return: None
        :rtype: None
        """
        dump = {
            conf.name: {
                "energy": conf.energy,
                "gsolv": conf.gsolv,
                "grrho": conf.grrho,
            }
            for conf in self.conformers
        }
        file.write_text(json.dumps(dump, indent=4))
